/*
 * Copyright 2003 LAAS-CNRS
 *
 *   LAAS-CNRS
 *   7, avenue du Colonel Roche
 *   F-31077 Toulouse Cedex 4 / France
 *
 * Author: Francois Taiani
 *
 * This file is part of OPSBrowser.
 * 
 * OPSBrowser is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * OPSBrowser is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with OPSBrowser; if not, write to the Free Software Foundation,
 * Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 * 
 */
/**
 * AbstractApplyToNodesCmd.java
 * 
 * Time-stamp: <2003-03-11 11:21:38 ftaiani>
 * @author Francois Taiani   <taiani@laas.fr>
 * CVS Info: $Id: AbstractApplyToNodesCmd.java,v 1.3 2003/06/13 16:38:22 ftaiani Exp $
 * 
 */

package laas.friends.cosmopen.commands;

import laas.friends.cosmopen.graphAnalysis.* ;
import java.util.* ;

/**
 * This class is an abstract class to implement commands
 * that select a set of vertices in a graph according
 * to a given patern and then repeadedly apply an
 * operation.
 * @author Francois Taiani   <taiani@laas.fr>
 * @version $Name: release-1-3-rc1 $ $Revision: 1.3 $
 * @since Version 0.1.2
 */
abstract public class AbstractApplyToNodesCmd extends CommandWithVariableArgs {

    private boolean shortOutputWanted = false ;

//     /**
//      * tell the class is all the nodes processed should be listed or
//      * only a short message. The default it a long output.
//      */
//     public void setShortOutput(boolean shortOuput) {
// 	shortOutputWanted = shortOuput ;
//     }

    public AbstractApplyToNodesCmd(String aName,int max,int min) {
	super(aName,max,min) ;
    } // EndConstructor AbstractApplyToNodesCmd

    protected DirectedGraphWithDict graphUnderFire ;

    /**
     * This methods computes the set of vertices that are removed
     * from the variable passed as second parameter from the 
     * condition expressed in the first parameter. It is abstract
     * in {@link AbstractRemove} and should be overwritten by
     * subclasses.
     */
    abstract protected Collection getNodeSToRemoveFromCondition
	( String condition ) ;

    protected int _nbOfRemovedNodes = 0 ;

    /**
     * the operation that is to be applied to the nodes selected. Should
     * be overwritten by subclasses. {@link #_nbOfRemovedNodes} should
     * be incremented in the overwritten method each time a node is removed.
     */
    abstract protected void applyCommandToNode(NamedVertex currentNamedVertex) 
	throws Exception ;

    /**
     * Is called by {@link #handleCall(String[])} to put the command
     * in the history. May be overwritten to change in particular the args
     * that are recorded. By default, we assume the command is of type
     * "<command> <pattern> <graphVariable>", i.e. only the second
     * argument is a graph variable.
     */
    protected void notifyChangeToHistory(String[] args) {
	GraphVariable.NotifyChangeOnVariable(args[2],args) ;
	GraphVariable.PutHistoryArgsInputOutput
	    ( args, 
	      new String[] {args[2]},
	      new String[] {args[2]}
	      );
    } // EndMethod notifyChangeToHistory

    /** ***************************************************}*
     * Validates the types of graph the command accepts.
     * This method is called by {@link #handleCall(String[] args)}
     * to validate the type of the graph on which the action
     * should take place. If the graph is not of the expected type
     * an {@link IllegalArgumentException} is thrown. The default
     * implementation does nothing, i.e. it accepts all type of graph.
     * To restrict the types of graph a command can be applied to
     * overload this method, test the type of {@link #graphUnderFire}
     * and throw an exception in case in does not match your
     * exceptation.
     */
    protected void validateOperandGraphType() throws IllegalArgumentException {} ;

    public boolean handleCall(String[] args) throws Exception {
	if ( ! doNecessaryVariableChecks(args,2,2)  ) return false ;
	if ( CheckIsBaseGraphAndErrorIfYes(args[2]) ) return false ;

	boolean executionOk = false ;
	_nbOfRemovedNodes = 0 ;
	try {
	    graphUnderFire = GraphVariable.GetGraphForName(args[2]) ;
	    validateOperandGraphType();
	    Collection result = getNodeSToRemoveFromCondition(args[1]) ;
	    if (result.isEmpty())
		println("No Matching Nodes in graph '"+args[2]+"'.") ;
	    else {

		// *** LOOP THROUGH NODES WITH MATCHING PATTERN

		java.util.Iterator iter = result.iterator() ;
		while (iter.hasNext()) {
		    NamedVertex currentNamedVertex = (NamedVertex)iter.next() ;
// 		    if (!shortOutputWanted)
// 			println(currentNamedVertex.getName()) ;
		    applyCommandToNode(currentNamedVertex) ;
		} // EndWhile Loop through result
		NamedVertex.PrintCollectionOfNamedVertex(result) ;
// 		if (shortOutputWanted)  System.out.print(_nbOfRemovedNodes+" nodes ") ;
// 		else                    System.out.print("   ") ;
		println("were removed from "+args[2]) ;
		notifyChangeToHistory(args) ;
		executionOk = true ;
	    } // EndElse there are matching nodes in passed graph
	} 
	catch (IllegalArgumentException except) {
	    println( "ERROR : " + except ) ;
	} // EndCatch
	return executionOk ;
    } // EndMethod handleCall

} // EndClass AbstractApplyToNodesCmd
