/*
 * Copyright 2003 LAAS-CNRS
 *
 *   LAAS-CNRS
 *   7, avenue du Colonel Roche
 *   F-31077 Toulouse Cedex 4 / France
 *
 * Author: Francois Taiani
 *
 * This file is part of OPSBrowser.
 * 
 * OPSBrowser is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * OPSBrowser is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with OPSBrowser; if not, write to the Free Software Foundation,
 * Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 * 
 */
/**
 * GraphVariable.java
 * 
 * Time-stamp: <2003-03-11 11:28:35 ftaiani>
 * @author Francois Taiani   <taiani@laas.fr>
 * CVS Info: $Id: GraphVariable.java,v 1.5 2007/09/18 18:23:41 ftaiani_normal Exp $
 * 
 */
package laas.friends.cosmopen.commands  ;

import laas.friends.cosmopen.graphAnalysis.* ;
import laas.friends.cosmopen.utils.MiscUtils ;
import laas.friends.cosmopen.interpreter.Command ;

import java.io.File      ;
import java.util.Set     ;
import java.util.HashSet ;
import java.util.Iterator;
import java.util.List    ;
import java.util.Vector  ;

/**
 * This class is used in the interpreter ({@link laas.friends.cosmopen.OPSBrowserInterpreter})
 * to store intermediary graph result. It basically consists of a 
 * hashtable and associates names (string) of variables with values.
 *
 * @author Francois Taiani   <taiani@laas.fr>
 * @version $Name: release-1-3-rc1 $ $Revision: 1.5 $
 * @since Version 0.0.3
 */
abstract public class GraphVariable {

    /**
     * The base graph when defined is accesible as normal
     * variable through the alias defined here.
     */
    public static String NAME_OF_BASEALIAS = "base" ;

    static private java.util.Map
	_VariableMap = new java.util.Hashtable() ;

    /**
     * sets the <b>global</b> context in which variables get their values.
     * No GraphVariable may be instantiated as long as this static method has
     * no been called.
     * This methods can't be called successively either. If the global
     * context of GraphVariables has to be changed, then {@link #ClearAllVariables() }
     * must be called first.
     * @see #ClearAllVariables()
     */
    static void SetBaseGraph(DirectedGraphWithDict aBaseGraph) {
	SetVariableToValue(NAME_OF_BASEALIAS,aBaseGraph) ;
    } // EndMethod setBaseGraph

    static DirectedGraphWithDict GetBaseGraph() {
	VariableEntry entry = 
	    (VariableEntry)GraphVariable._VariableMap.get(NAME_OF_BASEALIAS) ;
	return entry.graphValue ;
    } // EndMethod GetBaseGraph

    static boolean BaseGraphIsSet() {
	return (GraphVariable.VariableWithNameIsKnown(NAME_OF_BASEALIAS));
    } // EndMethod BaseGraphIsSet()

    /**
     * gets the value of tyhe variable with the passed name. 
     * If none is found, an exception is thrown.
     * @see #VariableWithNameIsKnown(String aName) 
     * @see #GetGraphForNameNoExp(String aName)
     */
    static DirectedGraphWithDict GetGraphForName(String aName) {
	VariableEntry result 
	    = (VariableEntry)GraphVariable._VariableMap.get(aName) ;
	if (result==null) 
	    throw new RuntimeException("No GraphVariable with name " + aName ) ;
	return result.graphValue ;
    } // EndMethod GetGraphForName

    /**
     * gets the value of tyhe variable with the passed name. 
     * If none is found, null is returned.
     * @see #VariableWithNameIsKnown(String aName) 
     * @see #GetGraphForName(String aName) 
     */
    static DirectedGraphWithDict GetGraphForNameNoExp(String aName) {
	VariableEntry result 
	    = (VariableEntry)GraphVariable._VariableMap.get(aName) ;
	if (result==null) 
	    return null ;
	return result.graphValue ;
    } // EndMethod GetGraphForName

    /**
     * defines a new variable with the passed name and value.
     */
    static public void SetVariableToValue
	(  String                aName,
	   DirectedGraphWithDict aValue ) 
    {
	// [CHANGE] ftaiani Mon Jul  2 2001 18:52:06
	// Constraint was relaxed because to complex to handle and enforce
	// consistently with optimization of load command.

// 	if (  aName.equals(NAME_OF_BASEALIAS) && 
// 	      BaseGraphIsSet() )
// 	    throw new RuntimeException
// 		( "Trying set a new BaseGraph for GraphVariables while "+
// 		  "old one still active. Use ClearAllVariables().") ;

	if ( !aName.equals(NAME_OF_BASEALIAS) && 
	     !BaseGraphIsSet() )
	    throw new RuntimeException
		( "Trying to instanciate a GraphVariable "+ 
		  "while no BaseGraph is set." ) ;

	// We look for a variable with the name aName.
	// If such a variable does not exist we instanciate
	// a new variable with this name.

	VariableEntry entryToModify ;
	entryToModify = (VariableEntry)GraphVariable._VariableMap.get(aName) ;

	if (entryToModify==null) { // no variable with this name
	    entryToModify = new VariableEntry(null) ;
	    GraphVariable._VariableMap.put (aName,entryToModify) ;
	} // EndIf no variable with this name
	
	entryToModify.graphValue = aValue ;

    } // EndMethod SetVariableToValue

    /**
     * returns a sorted map which maps variable names (String)
     * to their graph value 
     * ({@link laas.friends.cosmopen.graphAnalysis.DirectedGraphWithDict }).
     */
    static public java.util.SortedMap GetKnownVariables() 
    {
	// The problem here is to transform the internal mapping
	// GraphVariable._VariableMap between variable names and
	// composite VariableEntry's into a direct mapping between
	// variable names and the graph value of these variables.
	// For that we go through the mapping GraphVariable._VariableMap
	// and we reconstruct a new mapping, which is returned.

	java.util.SortedMap result = new java.util.TreeMap() ;
	java.util.Iterator  iter   
	    = GraphVariable._VariableMap.entrySet().iterator() ;

	// LOOP THROUGH ALL KNOWN VARIABLES

	while(iter.hasNext()) {
	    java.util.Map.Entry 
		currentMapEntry = (java.util.Map.Entry)iter.next() ;
	    VariableEntry
		currentVarEntry = (VariableEntry)currentMapEntry.getValue() ;
	    result.put(currentMapEntry.getKey(),currentVarEntry.graphValue) ;
	} // EndWhile Loop through known variable

	return result ;

    } // EndMethod GetKnownVariables
    
    /**
     * this methods returns the commands carried out on the given
     * variable. If the variable is not known, the behaviour is
     * not defined. The commands are retured as a list of strings.
     * The first string is the oldest command.
     */
    static public History GetHistoryForVariable(String aName) {
	// [CHANGE] fataiani Wed Jul  4 2001 10:51:04
	// We now implement an history function which takes the
	// interdependance between variables into account.
// 	VariableEntry entry 
// 	    = (VariableEntry)GraphVariable._VariableMap.get(aName) ;
// 	return entry.history ;

	Set varOfInterest = new HashSet() ;
	varOfInterest.add(aName);
	History result = new History() ;

	// [DEBUG] ftaiani Wed Jul  4 2001 12:20:25
// 	println("\nREVERSE LOOP ON _GlobalHistory\n") ;
// 	println("VarIns : "+varOfInterest );

	// *** REVERSE LOOP ON _GlobalHistory

	for( int i = _GlobalHistory.size()-1 ; i >= 0; i-- ) {
	    HistoryEntry currentEntry = (HistoryEntry)_GlobalHistory.get(i) ;
	    if (currentEntry.isRelevantForVarOfInterest(varOfInterest))
		{
		    result.add(0,currentEntry.command) ;

		    // [DEBUG] ftaiani Wed Jul  4 2001 12:20:25
// 		    println("Adding : "+currentEntry.command);
 		    currentEntry.modifyVarOfInterest(varOfInterest) ;
// 		    println("VarIns : "+varOfInterest );

		} // End currentEntry.isRelevantForVarOfInterest
	} // EndFor reverse loop on _GlobalHistory

//	println("\nReturning\n") ;

	return result ;

    } // EndMethod  GetHistoryForVariable

//  private static java.util.List _GlobalHistory = new java.util.Vector() ;
    private static java.util.List _GlobalHistory = new History() ;

    /**
     * This method returns a list of all past commands with
     * modification impact since the last clearall command.
     * This list is a copy of the real list, and can be modified
     * without problem.
     */
    static public History GetGlobalHistory() {
	
	History result = new History() ;

	// *** LOOP ON _GlobalHistory

	for( int i = 0 ; i < _GlobalHistory.size(); i++ ) {
	    HistoryEntry currentEntry = (HistoryEntry)_GlobalHistory.get(i) ;
	    result.add(currentEntry.command) ;
	} // EndFor reverse loop on _GlobalHistory

	return result ;
// 	return _GlobalHistory ;
    } // EndMethod GetGlobalHistory

    // [CHANGE] ftaiani Wed Jul  4 2001 12:23:55

//     /**
//      * this is used to update the history field of the opsbrowser
//      * interpreter.
//      * This method <b>must</b> called by each command that change
//      * the value of a variable, or of an external resource (like
//      * a file for instance).
//      * @param args           the arguments of the command. args[0] is the
//      *                       name of the comand.
//      */
//     static public void RegisterOnGlobalHistory(String[] args) {
// 	StringBuffer buffer = new StringBuffer() ;
// 	buffer.append(args[0]);
// 	for(int i=1; i<args.length; i++) {
// 	    buffer.append(' ').append(args[i]) ;
// 	} // EndFor i
// 	_GlobalHistory.add(buffer.toString()) ;
//     } // EndMethod RegisterOnGlobalHistory

    /**
     * this is used to update the history field of the opsbrowser
     * interpreter.
     * This method <b>must</b> called by each command that change
     * the value of a variable, or of an external resource (like
     * a file for instance).
     * @param args           the arguments of the command. args[0] is the
     *                       name of the comand.
     * @param  modifiedVar   names of the variable(s) whose value whose modified
     *                       by the command.
     * @param  inputVar      names of the variable(s) whose value influenced the
     *                       result of the command.
     */
    static public void PutHistoryArgsInputOutput
	(String[] args,
	 String[] inputVar, 
	 String[] modifiedVar) 
    {
	_GlobalHistory.add(new HistoryEntry(args,inputVar,modifiedVar)) ;
    } ; // EndMethod PutHistoryArgsInputOutput

    private static boolean inNotifyChangeOnVariable = false ; // Very base style

    /**
     * this is used to update the history field of the variable.
     * This method <b>must</b> called by each command that change
     * the value of a variable. The behaviour is not defined in
     * case the variable is not defined.
     * @param aVariableName  the name of the variable
     * @param args           the arguments of the command. args[0] is the
     *                       name of the comand.
     */
    static public void NotifyChangeOnVariable
	(String aVariableName, String[] args) 
    {
	VariableEntry entry 
	    = (VariableEntry)GraphVariable
	    ._VariableMap
	    .get(aVariableName) ;

	StringBuffer buffer = new StringBuffer() ;
	buffer.append(args[0]);
	for(int i=1; i<args.length; i++) {
	    buffer.append(' ').append(args[i]) ;
	} // EndFor i
	entry.history.add(buffer.toString()) ;

	if (entry.psFileToWriteTo!=null) 
	    UpdateBoundPSFile(entry) ;

    } // EndMethod NotifyChangeOnVariable

    private static void UpdateBoundPSFile(VariableEntry entry) 
    {
	inNotifyChangeOnVariable = true ;

	PrintGraphToPSFileWithDotWriter
	    ( entry.graphValue      ,
	      entry.psFileToWriteTo ,
	      entry.dotWriter       ) ;
	
	inNotifyChangeOnVariable = false ;

    } // EndMethod UpdateBoundPSFile

    /**
     * removes the passed variable from the table of the known
     * variable. Note that the underlying graph may further exists !!
     */
    static void DeleteVariable(String aName) {
	GraphVariable._VariableMap.remove(aName) ;
    } // EndMethod DeleteVariable

    /**
     * As for {@link #GetGraphForName(String aName) } except that in case
     * no variable already exists with the passed name, a new variable
     * with that name is created, and no exception if thrown. The new
     * variable is then initialized with a new empty graph which is returned. 
     * (i.e. <b>not</b> with the null reference.)
     */
    static DirectedGraphWithDict ProvideGraphValueForName(String aName) {
	VariableEntry resultEntry ;
	resultEntry = (VariableEntry)GraphVariable._VariableMap.get(aName) ;

	DirectedGraphWithDict result ;

	if (resultEntry==null) { // no variable with this name
	    result = GraphVariable.NewEmptyGraphSameTypeAsBase() ;
	    GraphVariable.SetVariableToValue(aName,result) ;
	} // EndIf no variable with this name
	else
	    result = resultEntry.graphValue ;

	return result ;
	
    } // EndMethod ProvideGraphValueForName

    /**
     * Clears the whole mapping between variables and their name (the variables
     * themselves are not cleared!). Set the global BaseGraph context to null.
     */
    static void ClearAllVariables() {
	GraphVariable._VariableMap  .clear() ;
	GraphVariable.ClearGlobalHistory() ;
    } // EndMethod ClearAllVariables

    static void ClearGlobalHistory() {
	GraphVariable._GlobalHistory.clear() ;
    }

    static boolean VariableWithNameIsKnown(String aName) {
	return GraphVariable._VariableMap.containsKey(aName) ;
    } // EndMethod VariableWithNameIsKnown

    static private java.util.Map mapPsFilesToVariables = new java.util.HashMap() ;

    private static void NotifyWriteToAFile(String aPsFile) {
	if (!inNotifyChangeOnVariable) {
	    UnbindVariableFromFile(aPsFile) ;
	    inNotifyChangeOnVariable = false ;
	}
	return ;

    } // EndMethod notifyWriteToAFileFromTarget

    public static void UnbindVariableFromFile(String file) {
	
	VariableEntry target = (VariableEntry)mapPsFilesToVariables.get(file) ;
	if (target!=null) {
	    mapPsFilesToVariables.remove(target.psFileToWriteTo) ;
	    target.psFileToWriteTo = null ;
	}

    } // EndMethod UnbindVariableFromFile

    /**
     * binds a variable to a ps file. Each change on the variable will
     * cause its content to be printed in the ps file. If the
     * passed name is the nulll string, or the refered
     * variable is not known, nothing is done (no exception).
     */
    static public void BindVariableToFileRawOrNot
	( String  aVarName, 
	  String  aPsFileName,
	  boolean inRawFormat )
    {
	if (aVarName==null) return ;
	if (aPsFileName!=null)
	    if (!aPsFileName.endsWith(".ps")) aPsFileName += ".ps" ;
	VariableEntry target 
	    = (VariableEntry)GraphVariable._VariableMap.get(aVarName) ;
	if (target==null) return ;

	target.psFileToWriteTo = aPsFileName ;
	mapPsFilesToVariables.put(aPsFileName,target) ;

	if (inRawFormat) target.dotWriter = target.graphValue.getRawWriter       () ;
	else             target.dotWriter = target.graphValue.getFormattingWriter() ;

	UpdateBoundPSFile(target) ;

    } // EndMethod BindVariableToFileAndDotWriter

    /**
     * same as {@link #PrintGraphToPSFile
     *	( DirectedGraphWithDict,String,String,String) PrintGraphToPSFile(
     *  sourceGraph, aFileName, pageSize, entireSize)} but without size
     * specifications.
     */
    static public void PrintGraphToPSFile
	(  DirectedGraphWithDict sourceGraph      ,
	   String                aFileName )
    {
	PrintGraphToPSFileWithPageSizeSizeAndComment
	    (sourceGraph,aFileName,null,null,null) ;
    } // EndMethod PrintGraphToPSFile

    static public void PrintGraphToPSFileWithDotWriter
	(  DirectedGraphWithDict sourceGraph      ,
	   String                aFileName        ,
	   DirectedGraphWithDict2DotOStream writer )
    {
	PrintGraphToPSFileWithPageSizeSizeAndCommentWithDotWriter
	    (sourceGraph,aFileName,null,null,null,writer) ;
    }

    static private boolean formatForScreen = true ;
    
    static public void setFormatForScreen(boolean newFormatForScreen) {
	formatForScreen = newFormatForScreen ;
    } // EndMethod setFormatForScreen


    static public void PrintGraphToPSFileWithPageSizeSizeAndComment
	(  DirectedGraphWithDict sourceGraph,
	   String                aFileName  ,
	   String                pageSize   , 
	   String                entireSize ,
	   String                comment    )
    {

	    DirectedGraphWithDict2DotOStream
		writer = (DirectedGraphWithDict2DotOStream)sourceGraph.getFormattingWriter() ;

	    writer.setFormatForScreen(formatForScreen) ;

	    PrintGraphToPSFileWithPageSizeSizeAndCommentWithDotWriter
		( sourceGraph,
		  aFileName  ,
		  pageSize   , 
		  entireSize ,
		  comment    ,
		  writer ) ;

	    writer.setFormatForScreen(true) ;

    }

    /**
     * print the content of a graph to a ps file.
     * In case something happen, this method prints error messages.
     */
    static private void PrintGraphToPSFileWithPageSizeSizeAndCommentWithDotWriter
	(  DirectedGraphWithDict sourceGraph,
	   String                aFileName  ,
	   String                pageSize   , 
	   String                entireSize ,
	   String                comment    ,
	   DirectedGraphWithDict2DotOStream writer )
    {
	if (aFileName==null) return ;
	if (!aFileName.endsWith(".ps")) aFileName += ".ps" ;

	try {

	    // We first ouput the value of the variable (which is
	    // a graph), in a intermediary file in dot format, using
	    // a DirectedGraphWithDict2DotOStream instance.
	    // Then we call the 'dot' utility in a separated process
	    // on this intermediary file to produce a ps output.
	    // (Process.waitFor())

	    File temporaryDotFile = File.createTempFile("opsbrowser", "tmp", new File(".")) ;
//  	    System.out.println("Writing to " + temporaryDotFile ) ;
	    java.io.OutputStream output 
		= new java.io.FileOutputStream(temporaryDotFile) ;
	    writer.setOutstream(output) ;
	    writer.setPageSizeAndEntireSize(pageSize,entireSize) ;
	    writer.setComment(comment) ;
	    DirectedGraphWithDict graphToWrite ;

// 	    if (sourceGraph instanceof CallGraph) {
// 		println("Converting call graph.") ;
// 		graphToWrite = callGraphPlugin.fromCallGraphToGraph((CallGraph)sourceGraph);
// 	    }
// 	    else {
// 		println("No Conversion needed.") ;
// 		graphToWrite = sourceGraph ;
// 	    }
// 	    // DEBUG
// 	    graphToWrite = sourceGraph ;

 	    writer.writeGraphInDotFormat(sourceGraph) ;
	    output.close() ;
	    NotifyWriteToAFile(aFileName) ;
//  	    System.out.println("Launching dot") ;
	    Runtime
		.getRuntime()
		.exec( "dot -Tps -o "+
		       aFileName +" "+
		       temporaryDotFile.getAbsolutePath() )
		.waitFor() ;

 	    temporaryDotFile.delete() ;
	    
	} catch (java.io.IOException ioExcept) {
	    Command.println( "IO ERROR: " + ioExcept) ;
// 	    ioExcept.printStackTrace() ;
	} catch (Exception except) {
	    Command.println( "INTERNAL ERROR: " + except ) ;
	    except.printStackTrace() ;
	} 
	
    } // EndMethod PrintGraphToPSFileWithPageSizeSizeAndComment

    /**
     * this class is used internally to manage the mapping
     * between names and values, and correctly handle
     * such information like history lists and the like.
     */
    private static class VariableEntry {
	java.util.List        history ;
	DirectedGraphWithDict graphValue ;
	/**
	 * when this variable is not null, each change
	 * on the variable will cause its content to be
	 * printed in a ps format in the file whose name
	 * is contained in the string. This variable
	 * is used to bind a variable to a file.
	 */
	String                           psFileToWriteTo = null ;
	DirectedGraphWithDict2DotOStream dotWriter = null ;
	VariableEntry(DirectedGraphWithDict aValue) {
	    this.graphValue   = aValue ;
	    this.history      = new java.util.Vector() ;
	} // EndConstructor
    } // EndInnerClass VariableEntry

    private static class HistoryEntry {
	private Set   _modifiedVariables = new HashSet() ;
	private Set   _inputVariable     = new HashSet() ;
	public String command ;
	
	public HistoryEntry( String[] args       ,
			     String[] inputVar   ,
			     String[] modifiedVar
			       ) 
	{
	    MiscUtils.PutArrayInCollection (modifiedVar, _modifiedVariables) ;    
	    MiscUtils.PutArrayInCollection (inputVar   , _inputVariable    ) ;

	    StringBuffer buffer = new StringBuffer() ;
	    buffer.append(args[0]);
	    for(int i=1; i<args.length; i++) {
		buffer.append(' ').append(args[i]) ;
	    } // EndFor i
	    command = buffer.toString() ;

	    // [DEBUG] ftaiani Wed Jul  4 2001 12:19:56
// 	    println("\nNew HistoryEntry") ;
// 	    println("  _inputVariable    :"+_inputVariable ) ;
// 	    println("  _modifiedVariables:"+_modifiedVariables) ;
// 	    println("\n") ;

	} // EndConstructor HistoryEntry

	/**
	 * return true if the {@link GraphVariable.HistoryEntry } modifies
	 * variables which are present in the passed {@link java.util.Set }
	 * <code>varOfInterest</code>.
	 * @param varOfInterest a Set of variable names ({@link String}s)
	 */
	public boolean isRelevantForVarOfInterest(Set varOfInterest ) {
	    Iterator i = _modifiedVariables.iterator() ;
	    while(i.hasNext()) {
		String nameModified = (String)i.next() ;
		if (varOfInterest.contains(nameModified)) return true ;
	    } // EndWhile Loop through _modifiedVariables
	    return false ; // No common variable found
	} // EndMethod isRelevantForVarOfInterest

	/**
	 * Modifies the passed {@link java.util.Set } of variable
	 * of interest accordingly to the action of the {@link GraphVariable.HistoryEntry }.
	 * Modified variables are removed and input variable are added.
	 * If a modified variable is an input as well, it remains
	 * in <code>varOfInterest</code>.
	 * @param varOfInterest a Set of variable names ({@link String}s)
	 */
	public void modifyVarOfInterest(Set varOfInterest) {
	    varOfInterest.removeAll(_modifiedVariables);
	    varOfInterest.addAll   (_inputVariable    );
	} // EndMethod getInputVar

    } // EndInnerClass HistoryEntry

    static DirectedGraphWithDict NewEmptyGraphSameTypeAsBase() {
	    if (BaseGraphIsSet()) return GetBaseGraph().getEmptyGraphOfSameType() ;
	    else                  return new DirectedGraphWithDictImpl() ;
    } // EndMethod NewEmptyGraphSameTypeAsBase

//     static String GetNameForEventID(int eventID) {

// 	CallGraph base = (CallGraph)GraphVariable.GetBaseGraph() ;
// 	NamedVertex borderEvent = base.getEventForID(eventID);
// 	return GetNameProtectIfNull(borderEvent) ;

//     } // EndMethdo GetNameForEventID

    static String GetMostRecentEventNameBeforeEventID(int eventID) {

	CallGraph base = (CallGraph)GraphVariable.GetBaseGraph() ;
	NamedVertex borderEvent = base.getMostRecentEventBeforeID(eventID);
	String name ;
	if (borderEvent!=null) name = borderEvent.getName() ;
	else                   name = "No event for ID "+ eventID ;
	return name ;

    } // EndMethdo GetNameForEventID

} // EndClass GraphVariable

