/*
 * Copyright 2004 Francois Taiani
 *
 * Author: Francois Taiani
 *
 * This file is part of OPSBrowser.
 * 
 * OPSBrowser is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * OPSBrowser is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with OPSBrowser; if not, write to the Free Software Foundation,
 * Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 * 
 */
/**
 * BreakpointTraceExtractor.java
 * 
 * Time-stamp: <>
 * @author Francois Taiani   <f.taiani@computer.org>
 * $Id: BreakpointTraceExtractor.java,v 1.4 2005/03/11 19:53:45 ftaiani Exp $
 * 
 */

package laas.friends.cosmopen.extractor ;

import java.io.* ;
import java.util.* ;
import laas.friends.cosmopen.utils.* ;

import com.sun.jdi.*;
import com.sun.jdi.request.*;
import com.sun.jdi.event.*;
import com.sun.jdi.connect.* ;

/**
 * This class extracts a stack trace sequence from a java program
 * using breakpoints. It relies on the 
 * @author Francois Taiani   <f.taiani@computer.org>
 */
public class BreakpointTraceExtractor {

    private PrintStream         myOutStream = System.out ;
    private LaunchingConnector  myConnector ;
    private VirtualMachine      myTargetJVM ;
    private EventRequestManager myEventRManager ;

    private boolean myJVMisConnected = false ;
    private Map     classPrepareRequestToBreakpointMethod = new HashMap() ;
    private boolean javaLangThreadStartActivated = false ;

    // ***********************************************************

    public BreakpointTraceExtractor(String[] targetVMOptionsAndStartingClass)
        throws IOException, 
               IllegalConnectorArgumentsException,
               VMStartException
    {
        sortOutMainClassFromJVMOptions(targetVMOptionsAndStartingClass) ;
        initialize() ;

    } // EndConstructor BreakpointTraceExtractor

    public BreakpointTraceExtractor(String startingClassAndArgs,
                                    String targetVMOptions) 
        throws IOException, 
               IllegalConnectorArgumentsException,
               VMStartException
    {
        myMainClass        = startingClassAndArgs ;
        myJVMOptionsString = targetVMOptions ;
        initialize() ;

    } // EndConstructor BreakpointTraceExtractor
     
    /**
     * Common stuff for constructors.
     */
    private void initialize() 
        throws IOException, 
               IllegalConnectorArgumentsException,
               VMStartException
    {
        myConnector = Bootstrap.virtualMachineManager().defaultConnector() ;

        Map jvmArgumentS = myConnector.defaultArguments();

        Connector.Argument mainAndOptions 
            = (Connector.Argument)jvmArgumentS.get("main");

        mainAndOptions.setValue(myMainClass);

	System.err.println(mainAndOptions);

        if (myJVMOptionsString!=null) {
            Connector.Argument jvmOptions
                = (Connector.Argument)jvmArgumentS.get("options");
            jvmOptions.setValue(myJVMOptionsString);
        } 

        myTargetJVM      = myConnector.launch(jvmArgumentS);
        myJVMisConnected = true ;
        myEventRManager  = myTargetJVM.eventRequestManager();

        ThreadStartRequest threadStartRequest =
            myEventRManager.createThreadStartRequest() ;
        
        threadStartRequest.setSuspendPolicy(EventRequest.SUSPEND_ALL);
        threadStartRequest.enable();

        ThreadDeathRequest threadDeathRequest =
            myEventRManager.createThreadDeathRequest() ;
        
        threadDeathRequest.setSuspendPolicy(EventRequest.SUSPEND_ALL);
        threadDeathRequest.enable();

    } // EndMethod initialize

    public boolean targetJVMisConnected() {
        return myJVMisConnected ;
    } // EndMethod isTargetJVMConnected

    // ***********************************************************

    ExceptionRequest allExceptionRequest = null ;

    /**
     * switch on or off the recording of exception in the target JVM. 
     */
    public void activateExceptionExtraction(boolean doActivate) {

        if (doActivate && (allExceptionRequest == null)) {

            allExceptionRequest = 
                myEventRManager.createExceptionRequest(null, true, true); 
            allExceptionRequest.addClassExclusionFilter("java.net.URLClassLoader*");
            allExceptionRequest.addClassExclusionFilter("java.lang.ClassLoader*");
            allExceptionRequest.addClassExclusionFilter("java.security.SecureClassLoader*");
            allExceptionRequest.setSuspendPolicy(EventRequest.SUSPEND_ALL);
            allExceptionRequest.enable();
        
        } // EndIf activation

        if (!doActivate && (allExceptionRequest != null)) {
            myEventRManager.deleteEventRequest(allExceptionRequest);
            allExceptionRequest = null ;

        } // EndIf desactivation

    } // EndMethod activateExceptionExtraction

    // ***********************************************************

    public void setBreakPoint(String className, String methodName) {

//             System.err.println("ClassPrepareRequest for "+className );

        List listOfClassesForName = myTargetJVM.classesByName(className) ;

        if(!listOfClassesForName.isEmpty()) {
            // CLASS IS KNOWN. IMMEDIATE BREAKPOINT ACTIVATION
            for(Iterator i = listOfClassesForName.iterator() ; i.hasNext() ; ) {
                ClassType currentClass = (ClassType)i.next() ;
                setBreakPointRequest(currentClass, methodName) ;
            } // EndFor
        }
        else { 
            // CLASS NOT KNOW. WE DELAY BREAKPOINT ACTIVATION
	    System.err.println("Delayed breakpoint activation");
            ClassPrepareRequest classPrepReq 
                = myEventRManager.createClassPrepareRequest();
            classPrepReq.addClassFilter(className);
            classPrepReq.setSuspendPolicy(EventRequest.SUSPEND_ALL);
            classPrepReq.enable();
        
            classPrepareRequestToBreakpointMethod.put(classPrepReq,methodName) ;
        }

    } // EndMethod setBreakPoint
    
    // ***********************************************************

    /**
     * Breakpoints can't be set on native methods, at least with
     * Sun'JVM, but MethodEntryRequest can, so we try to emulate this kind of behavior
     * for native methods.
     */
    private Map entryRequestToMethodForNatEmulation = new HashMap() ;

    private void addBreakPointEmulation(Method someMethod) {

        if (entryRequestToMethodForNatEmulation.containsValue(someMethod))
            return ;

        MethodEntryRequest newMethodEntryRqst 
            = myEventRManager.createMethodEntryRequest();

        newMethodEntryRqst.addClassFilter(someMethod.declaringType().name())  ;
        newMethodEntryRqst.setSuspendPolicy(EventRequest.SUSPEND_ALL) ;
        newMethodEntryRqst.enable();

        entryRequestToMethodForNatEmulation.put(newMethodEntryRqst,someMethod) ;

    } // EndMethod addBreakPointEmulation

    // ***********************************************************

    private void setBreakPointRequest(ClassType classType, 
                                      String    breakpointMethodName) {

        List listOfMethods = classType.methodsByName(breakpointMethodName);

        if (listOfMethods.isEmpty()) {
            System.err.println("Warning : Could not find method "+
                               classType.name()+"."+
                               breakpointMethodName) ;
        }

        for(Iterator i = listOfMethods.iterator() ; i.hasNext() ; ) {
            Method   breakMethod = (Method)i.next();
            if (breakMethod.isNative()) { 
                // Breakpoint are not allows on native method with
                // Sun's JVM. We try to emulate them with
                // MethodEntryRequest.
                addBreakPointEmulation(breakMethod) ;
            } else {
		System.err.println("New breakpoint request @"+breakMethod);
                BreakpointRequest brRequest = 
                    myEventRManager.createBreakpointRequest(breakMethod.location());
                brRequest.setSuspendPolicy(EventRequest.SUSPEND_ALL);
                brRequest.enable();
            }
            
        } // EndFor

    } // EndMethod setBreakPointRequest

    // ***********************************************************

    private Collection setOfMthdEntryReqstFor = new HashSet() ;

    public void setBreakPointExpression(String classPattern) {

        MethodEntryRequest newMethodEntryRqst ;

        newMethodEntryRqst = myEventRManager.createMethodEntryRequest();
        newMethodEntryRqst.addClassFilter(classPattern)  ;
        newMethodEntryRqst.setSuspendPolicy(EventRequest.SUSPEND_ALL) ;
        newMethodEntryRqst.enable();
        
        setOfMthdEntryReqstFor.add(newMethodEntryRqst) ;

    } // EndMethod setBreakPointExpression

    // ***********************************************************

    private boolean hasSessionBeenInitialized = false ;

    private void initCollectSessionIfNeed() {
        if (hasSessionBeenInitialized) return ;

        hasSessionBeenInitialized = true ;
        myOutStream.println("<traceSet threadCreation=\"java.lang.Thread::start\" >") ;
//         System.err.println ("<traceSet threadCreation=\"java.lang.Thread::start\" >") ;
        myTargetJVM.resume() ;
        
        eventQueue    = myTargetJVM.eventQueue();
        
        while (eventIterator==null) {
            try {
                eventSet      = eventQueue.remove() ;
                eventIterator = eventSet.eventIterator();
            } catch (InterruptedException e) {}
        }
    } // EndMethod initCollectSessionIfNeed

    // ***********************************************************

    private EventIterator eventIterator = null ;
    private EventQueue    eventQueue    = null ;
    private EventSet      eventSet      = null ;

    private Event getNextEvent() {

        while (!eventIterator.hasNext()) {
            try {
                eventSet.resume();
                eventSet = eventQueue.remove();
		eventIterator = eventSet.eventIterator();
            } catch (InterruptedException e) {}
        }
        return eventIterator.nextEvent() ;

    } // EndMethod getNextEvent

    // ***********************************************************

    private boolean lastEventGeneratedATrace = false ;
    private String  lastClassHit  = null ;
    private String  lastMethodHit = null ;

    public void collectEventTillGivenBreakPoint(String className,
                                                String methodName) {

        initCollectSessionIfNeed() ;
        boolean weHaveFoundTheBreakPointWeLookFor = false ;

        while(!weHaveFoundTheBreakPointWeLookFor && myJVMisConnected) {
            lastEventGeneratedATrace = false ;
            while (myJVMisConnected && !lastEventGeneratedATrace) {
                handleNextEvent() ; 
            }
//             System.err.println("lastClassHit : "+lastClassHit );
//             System.err.println("lastMethodHit: "+lastMethodHit);
            weHaveFoundTheBreakPointWeLookFor =
                className .equals(lastClassHit ) &&
                methodName.equals(lastMethodHit) ;

        } // EndWhile !weHaveFoundTheBreakPointWeLookFor

    } // EndMethod collectEventTillGivenBreakPoint

    public void collectEventTillNextBreakPoint() {

        initCollectSessionIfNeed() ;

        while (myJVMisConnected && !lastEventGeneratedATrace)

            handleNextEvent() ;

    } // EndMethod collectEventTillNextBreakPoint

    // ***********************************************************

    public void collectEventTillVMDeath() {

	//	System.err.println("Starting collectEventTillVMDeath");
        initCollectSessionIfNeed() ;
	//	System.err.println("after initCollectSessionIfNeed");

        EventQueue queue = myTargetJVM.eventQueue();

	//	System.err.println("after myTargetJVM.eventQueue");

        while (myJVMisConnected) {
            handleNextEvent() ;
        } // EndWhile

    } // EndMethod collectEventTillVMDeath

    // ***********************************************************

    private void handleNextEvent() {

	Event event ;

	try{
	    event = getNextEvent() ;
	}  catch (VMDisconnectedException e) { 
	    // Documentation reads "This will always be preceded
	    // by a VMDisconnectEvent." So no need to be caught.
	    
	    // (Francois Taiani) Fri Mar 11 2005 18:44:42:
	    // Well in practice it does not always on buggy implementations
	    // so we need to do something anyway
	    if (myJVMisConnected)
		reactToVMDisconnectEvent();
	    return ; // exiting while loop returning
	}

        lastEventGeneratedATrace = false ;

	System.err.println("handleNextEvent : "+ event) ;
        
        if        (event instanceof   ExceptionEvent) {
            reactToExceptionEvent (  (ExceptionEvent)event);
        } 
        else if (event instanceof     ThreadStartEvent) {
            if (javaLangThreadStartActivated)
                getBuilderForThreadID
                    (((ThreadStartEvent)event).thread())
                    .reactToThreadStartEvent();
        } 
        else if (event instanceof     ThreadDeathEvent) {
            if (javaLangThreadStartActivated)
                getBuilderForThreadID
                    (((ThreadDeathEvent)event).thread())
                    .reactToThreadDeathEvent();
        } 
        else if (event instanceof          LocatableEvent) {
            reactToLocatableEvent   ((LocatableEvent)event);
        }
        else if (event instanceof     ClassPrepareEvent) {
            reactToClassPrepareEvent((ClassPrepareEvent)event);
        } 
        else if (event instanceof     VMDisconnectEvent) {
            reactToVMDisconnectEvent();
        }
    } // EndMethod handleNextEvent

    // ***********************************************************

    private void reactToVMDisconnectEvent() {
        myJVMisConnected=false ;
        myOutStream.println("</traceSet>") ;
//         System.err.println ("</traceSet>") ;

    }

    // ***********************************************************

    /**
     * Set the breakpoint that are registered for a given class when
     * this class has been reported loaded.
     */
    private void reactToClassPrepareEvent(ClassPrepareEvent event)  {

        String breakpointMethodName = 
            (String)classPrepareRequestToBreakpointMethod.get(event.request());

        if (breakpointMethodName==null)
            throw new RuntimeException("INTERNAL ERROR : Intecepted class not as expected.") ;

	setBreakPointRequest((ClassType)event.referenceType(), breakpointMethodName) ;

        myEventRManager.deleteEventRequest(event.request());
        classPrepareRequestToBreakpointMethod.remove(event.request()) ;

    } // EndMethod reactToClassPrepareEvent

    // ***********************************************************

    private void reactToLocatableEvent(LocatableEvent event) {

        if (event instanceof     MethodEntryEvent) {
            reactToMethodEntryEvent( (MethodEntryEvent)event) ;
        } 
        else if (event instanceof     BreakpointEvent) {
            ThreadBuilder thrBuilder = getBuilderForThreadID(event.thread()) ;
            thrBuilder.printXMLStackTrace() ;
        }

    } // EndMethod reactToLocatableEvent

    // ***********************************************************

    /**
     * Here we try to emulate breakpoint behavior for native
     * methods by using MethodEntryEvent.
     */
    private void reactToMethodEntryEvent(MethodEntryEvent event) {

        if (setOfMthdEntryReqstFor.contains(event.request())) {
            // This is the standard MethodEntryEvent, no emulation here.
            ThreadBuilder thrBuilder = getBuilderForThreadID(event.thread()) ;
            thrBuilder.printXMLStackTrace() ;
            return ;
        }

        // We're emulating breakpoint so we only want the method
        // the request is registered with.

        Method registeredMethod = 
            (Method)entryRequestToMethodForNatEmulation.get(event.request()) ;

        if (registeredMethod==event.method()) {
            ThreadBuilder thrBuilder = getBuilderForThreadID(event.thread()) ;
            thrBuilder.printXMLStackTrace()  ;
        }

    } // EndMethod reactToMethodEntryEvent

    // ***********************************************************

    private void reactToExceptionEvent(ExceptionEvent event) {

        // If the exception was caught in the same
        // frame as it was thrown, we don't record it.

        Location        catchLocation = event.catchLocation() ;
        ThreadReference thread = event.thread() ;

        try {

             if (catchLocation!=null) {
                 Method catchMethod = catchLocation.method() ;
                 if (catchMethod!=null)
                     if (catchMethod.equals(thread.frame(0).location().method()))
                         return ;
             } // EndIf

        } catch (IncompatibleThreadStateException e) {
            e.printStackTrace() ;
            return ;
        }

        // The exception was caught outside of its originating
        // frame, or it was uncaught. We record it.

        getBuilderForThreadID(thread).printXMLExceptionStackTrace(event);

        //         System.err.println("Exception: " + event.exception() + 
        //                            " catch: " + event.catchLocation());
        
    } // EndMethod reactToExceptionEvent

    // ***********************************************************

    public void setOutputStream(OutputStream out) {
        myOutStream = new PrintStream(out) ;
    } // EndMethod setOutputStream
    
    // ***********************************************************

    private String packageRestriction = null ;
    
    public void restrictTraceToPackage(String packageName) {
        packageRestriction = packageName ;
    } // EndMethod restrictTraceToPackage

    // ***********************************************************

    private String myMainClass ;
    private String myJVMOptionsString   ;

    /**
     * Sorts out the options for the target JVM from the globus entry
     * class and its (possible) own options.
     */
    private void sortOutMainClassFromJVMOptions(String[] args) {

        StringBuffer optionsBuf   = new StringBuffer();
        StringBuffer mainClassBuf = new StringBuffer();

        int inx ;

        // LOOP THROUGH MAIN ARGS : Args for target JVM

        for (inx=0 ; (inx < args.length) ; inx++) {

	    String arg = args[inx];

            if (arg.charAt(0) != '-') break ; // This means we've found the main class

            optionsBuf.append(arg);
            optionsBuf.append(' ');

            if (arg.equals("-classpath") || arg.equals("-cp")) {
                inx++ ;
                if (!(inx < args.length))
                    throw new RuntimeException("-classpath needs an argument.") ;
                arg = args[inx];
                if (arg.charAt(0) == '-')
                    throw new RuntimeException("-classpath needs an argument.") ;
                optionsBuf.append(arg);
                optionsBuf.append(' ');
            } // EndIf "-classpath"

        } // EndFor LOOP THROUGH MAIN ARGS

       // LOOP THROUGH REMAING ARGS : Main class and its (possible) own options

        for (; (inx < args.length) ; inx++) {
            mainClassBuf.append(args[inx]);
            mainClassBuf.append(' ');

        } // EndFor LOOP THROUGH REMAINING ARGS

        myMainClass        = mainClassBuf.toString() ;
        myJVMOptionsString = optionsBuf  .toString() ;
        
    } // EndMethod sortOutMainClassFromJVMOptions

    // ***********************************************************
    // ***********************************************************

    private ThreadBuilder getBuilderForThreadID(ThreadReference someThread) {
        ThreadBuilder res = (ThreadBuilder)rawThreadIDtoThrdBuilder.get(someThread);
        if (res==null) 
            res = new ThreadBuilder(someThread) ;
        return res ;        
    } // EndMethod getBuilderForThreadID

    private int threadCount = 0 ;
    private Map rawThreadIDtoThrdBuilder = new HashMap() ;

    /** 
     * is responsible for the management of calls related to one given
     * thread.
     */
    private class ThreadBuilder {
        public String      niceThreadID ;
        public ThreadReference thread ;
 
        public ThreadBuilder(ThreadReference someThread) {
            threadCount++ ;
            niceThreadID = Integer.toString(threadCount) ;
            rawThreadIDtoThrdBuilder.put(someThread,this);
            thread = someThread ;
        } // EndConstructor ThreadBuilder()

        // ***********************************************************

        private void reactToThreadDeathEvent() {
            myOutStream.println("  <end thread=\""+niceThreadID+"\" />") ;
//             System.err.println ("  <new thread=\""+niceThreadID+"\" />") ;            
        } // EndMethod reactToThreadDeathEvent

        private void reactToThreadStartEvent() {
            myOutStream.println("  <new thread=\""+niceThreadID+"\" />") ;
//             System.err.println ("  <new thread=\""+niceThreadID+"\" />") ;            
        } // EndMethod reactToThreadStartEvent

        private void printXMLStackTrace() {
            printXMLStackTraceStart() ;
            printXMLStackTraceContent() ;
            printXMLStackTraceEnd() ;
        } // EndMethod printXMLStackTrace

        private void printXMLExceptionStackTrace(ExceptionEvent exceptionEvt) {
            printXMLStackTraceStart() ;
            printXMLCallElement(exceptionEvt.exception().referenceType().name(),"throw") ;
            printXMLStackTraceContent() ;
            printXMLStackTraceEnd() ;            
        } // EndMethod printXMLExceptionStackTrace

        private void printXMLStackTraceStart() {
            myOutStream.println("  <trace thread=\""+niceThreadID+"\" >") ;
//             System.err.println ("  <trace thread=\""+niceThreadID+"\" >") ;
        } // EndMethod printXMLStackTraceStart

        private void printXMLStackTraceEnd() {
                myOutStream.println("  </trace>") ;
//                 System.err.println ("  </trace>") ;
                lastEventGeneratedATrace = true ;
        } // EndMethod printXMLStackTraceEnd

        private void printXMLCallElement(String className, String methodName) {
            StringBuffer callString = new StringBuffer("    <call entity=\""); 
            callString.append(className) ;
            callString.append("\" method=\"") ;
            callString.append(MiscUtils.makeXMLCompliant(methodName)) ;
            callString.append("\" />") ;
            myOutStream.println(callString) ;
            //                     System.err.println (callString) ;
        } // printXMLCallElement

        private void printXMLStackTraceContent() 
        {

            if (!thread.isSuspended()) {
                System.err.println("Warning : Trying to generate a trace for a "+
                                   "thread that is not suspended ! Skipping.") ;
                return ;
            }

            try {


                boolean firstFrame = true ;

                for(Iterator frameIdx = thread.frames().iterator() ; 
                    frameIdx.hasNext() ; ) {

                    StackFrame currentFrame = (StackFrame)frameIdx.next() ;
                    Method     currentMethod= currentFrame.location().method() ;
                    String     className    = currentMethod.declaringType().name() ;
                    String     methodName   = currentMethod.name() ;

                    if (packageRestriction!=null)
                        if (!className.startsWith(packageRestriction)) continue ;

                    if (className.equals("java.lang.Thread"))
                        if (methodName.equals("start"))
                            javaLangThreadStartActivated = true ;

                    if (firstFrame) {
                        lastClassHit  = className ;
                        lastMethodHit = methodName ;
                        firstFrame    = false ;
                    }
                    printXMLCallElement(className,methodName) ;

                } // EndFor

            } catch(IncompatibleThreadStateException e) {
                throw new RuntimeException("INTERNAL ERROR : IncompatibleThreadStateException") ;
            }

        } // EndMethod printXMLStackTrace

    } // EndClass ThreadBuilder

} // EndClass BreakpointTraceExtractor

