/*
 * Copyright 2003 LAAS-CNRS
 *
 *   LAAS-CNRS
 *   7, avenue du Colonel Roche
 *   F-31077 Toulouse Cedex 4 / France
 *
 * Author: Francois Taiani
 *
 * This file is part of OPSBrowser.
 * 
 * OPSBrowser is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * OPSBrowser is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with OPSBrowser; if not, write to the Free Software Foundation,
 * Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 * 
 */
/**
 * CallGraph.java
 * 
 * Time-stamp: <2003-03-07 22:18:02 ftaiani>
 * @author Francois Taiani   <taiani@laas.fr>
 * CVS Info: $Id: CallGraph.java,v 1.16 2007/09/18 18:23:42 ftaiani_normal Exp $
 * 
 * @version $Name: release-1-3-rc1 $ $Revision: 1.16 $
 * @since Version 0.1.1
 */

package laas.friends.cosmopen.graphAnalysis  ;

import java.util.* ;
import laas.friends.cosmopen.utils.MiscUtils ;

/**
 * This class stores a call graph that takes into account the
 * causality between call events.  For instance if A calls B calls C,
 * and X calls B calls Y, then a back track from C would yield {A,B,X}
 * in a normal graph, though the call on C is not causally connected
 * to the call from X on B. The class <code>CallGraph</code> solves
 * that problem.<P>
 *
 * A {@link CallGraph} also manage the relative order of the
 * invocations ("events") it stores using "EventIDs".
 *
 * A <code>CallGraph</code> also maintain some information on its
 * internal structure in order to speed its construction from a trace
 * set (see {@link XMLTraceSet2CallGraphReader}).  This information
 * takes the form of extra "shadow" objects that are associated with
 * each {@link NamedVertex node} present in the graph. These "shadow"
 * objects are implemented by {@link ChildrenRegistry}, and keeps
 * structural information regarding the children of a given node
 * <i>within</i> the graph.  This is the reason why this information
 * is not added to each {@link NamedVertex}.
 */
public class CallGraph 
extends DirectedGraphWithDictImpl {

    /**
     * This variable associates each vertex of the {@link CallGraph}
     * with its corresponding {@link ChildrenRegistry}.
     * @see #topChildrenMap
     */
    protected TreeMap           mapOfChildrenRegistry    = null ;

    /** See {@link #mapOfChildrenRegistry} */
    protected ChildrenRegistry  topChildrenMap ;

    /**
     * This is a trick to handle vertices that are roots
     * (no parent) in the same way as normal vertices.
     * @see #mapOfChildrenRegistry
     */
    public final NamedVertex hiddenTopVertex ;

    protected SortedMap _mapEventIDToVertex ;

    /**
     * Is used to assign event order ID to new nodes in the call graph.
     */
    public int greatestEventIDInGraph ;

    protected boolean underConstructionFromTraceSet ;

    /** **************************************************** */
    public CallGraph() {
	topChildrenMap      = null ;
	hiddenTopVertex     = new NamedVertex("") ;
	_mapEventIDToVertex = new TreeMap() ;
	greatestEventIDInGraph        = -1 ;
	underConstructionFromTraceSet = false ;
	getPropBag().setPropertyToValue("type","callGraph") ;
    } // EndConstructor

    /** 
     * Copy constructor.
     */
    public CallGraph(CallGraph toBeCopied) {
	super(toBeCopied);
	topChildrenMap      = toBeCopied.topChildrenMap ;
	hiddenTopVertex     = toBeCopied.hiddenTopVertex ;
	_mapEventIDToVertex = new TreeMap(toBeCopied._mapEventIDToVertex) ;
	greatestEventIDInGraph        = toBeCopied.greatestEventIDInGraph ;
	underConstructionFromTraceSet = toBeCopied.underConstructionFromTraceSet ;
    } // EndConstructor copy constructor

    public Object clone() {
	return new CallGraph(this) ;
    } // EndMethod clone

    /** ****************************************************
     * Finished the construction of a CallGraph that
     * has been started using {@link #GetCallGraphUsingChildrenRegistry()}.
     */
    public void switchOffUseOfChildrenRegistry() {
	underConstructionFromTraceSet = false ;
	// We forget anything we may now about ChildrenRegistry
	topChildrenMap            = null ;
	mapOfChildrenRegistry  = null ;
    } // EndMethod 

    /** ****************************************************
     * Returns a {@link CallGraph} that is prepared to maintain
     * {@link ChildrenRegistry} information. This is needed
     * when constucting a {@link CallGraph} from a set of traces.
     * @see XMLTraceSet2CallGraphReader
     * @see #switchOffUseOfChildrenRegistry()
     */
    public static CallGraph GetCallGraphUsingChildrenRegistry() {
	CallGraph result = new CallGraph() ;
	result.initChildrenRegistryStuff() ;
	return result ;
    } // EndMethod GetCallGraphUsingChildrenRegistry


    /** **************************************************** */
    public void initChildrenRegistryStuff() {
	topChildrenMap            = new ChildrenRegistry() ;
	mapOfChildrenRegistry  = new TreeMap() ;
	mapOfChildrenRegistry.put(hiddenTopVertex,topChildrenMap) ;
	underConstructionFromTraceSet = true ;
    } // EndMethod InitChildrenRegistryStuff

    /** **************************************************** */
    public void add(salvo.jesus.graph.Vertex aNewNamedVertex) {

	NamedVertex newNamedVertex = (NamedVertex)aNewNamedVertex ;
	try {
	    if (doesContainVertex(newNamedVertex)) return ;
 	} catch(NamedVertexConflictException e) { return ; } // We're lazy on properties
	super.add(newNamedVertex) ;
	if (underConstructionFromTraceSet) {
	    topChildrenMap.addChild((NamedVertex)newNamedVertex) ;
	    mapOfChildrenRegistry.put(newNamedVertex,new ChildrenRegistry() ) ;
	}
	notifyVertexGotEventID(newNamedVertex) ;
    } // EndMethod add


    /** **************************************************** */
    public void notifyVertexGotEventID(NamedVertex aVertex) {
	String eventID = aVertex.getValueForProperty("eventOrder") ;
	if (eventID!=null) 
	    _mapEventIDToVertex.put(new Integer(eventID),aVertex) ;
    }
   

    /** **************************************************** */
    public void remove(salvo.jesus.graph.Vertex v)
	throws java.lang.Exception 
    {
	// When a vertex is removed, removeEdge is also called for all
	// its connected edges, so we don't need to do any bookkeeping
	// about its children here.
	super.remove(v) ;
	if (underConstructionFromTraceSet)
	    mapOfChildrenRegistry.remove(v) ;
	String eventID = ((NamedVertex)v).getValueForProperty("eventOrder") ;
	if (eventID!=null) 
	    _mapEventIDToVertex.remove(new Integer(eventID)) ;
    } // EndMethod remove


    /** **************************************************** */
    public ChildrenRegistry getChildrenRegistryForVertex(NamedVertex aCallNode) {
	if (!underConstructionFromTraceSet)
	    throw new RuntimeException
                ("INTERNAL ERROR: getChildrenRegistryForVertex "+
                 "was called on a CallGraph that is not under "+
                 "construction from a set of traces.") ;
	ChildrenRegistry 
            result = (ChildrenRegistry)mapOfChildrenRegistry.get(aCallNode) ;
	return (ChildrenRegistry)mapOfChildrenRegistry.get(aCallNode) ;
    }

    /** **************************************************** */
    public void removeEdge(salvo.jesus.graph.Edge e)
                throws java.lang.Exception
    {
	NamedVertex parent = (NamedVertex)((NamedEdge)e).getSource() ;
	NamedVertex child  = (NamedVertex)((NamedEdge)e).getSink  () ;
	if (underConstructionFromTraceSet)
	    getChildrenRegistryForVertex(parent).removeChild(child) ;
	super.removeEdge(e) ;
	if (underConstructionFromTraceSet)
	    if (getIncomingEdges(child).isEmpty()) {
		topChildrenMap.addChild(child) ;
	} // EndIf child.getIncomingEdges().empty()

    } // EndMethod removeEdge

    /** 
     * returns the parent of the node. (null if the node is a root.)
     */
    public NamedVertex getParent(NamedVertex anEvent) {
	NamedEdge filiationLink = getParentEdge(anEvent) ;
	if (filiationLink==null) return null ;
	return (NamedVertex)filiationLink.getSource() ;
    } // EndMethod getParent

    /** ****************************************************
     * a call graph is made of trees, so a given node has
     * at most one incoming edge. This method returns it (or null 
     * if the vertex is a root).
     */
    public NamedEdge getParentEdge(NamedVertex anEvent) {
	NamedVertex theRealEvent = getNodeWithName(anEvent.getName());
	if (theRealEvent==null) return null ;
	Collection      previousEdges = getIncomingEdges(theRealEvent) ;
	if (previousEdges.size()>1)
	    new RuntimeException
		    ( "A node in a call graph has more that one parent!!\n" +
		      theRealEvent +" : \n" + previousEdges ) ;
	if (previousEdges.size()==1) return (NamedEdge)previousEdges.iterator().next();
	return null ;
    } // EndMethod getParentEdge


    /** **************************************************** */
    public void addEdge(salvo.jesus.graph.Edge e) throws java.lang.Exception {
	NamedEdge   addedNamedEdge = (NamedEdge)e ;
	NamedVertex callee         = (NamedVertex)addedNamedEdge.getSink() ;
	NamedEdge   parentEdge     = getParentEdge(callee) ;

	if (parentEdge!=null) // We must do something to prevent an
                              // event having 2 parents
	    removeEdge(parentEdge) ;
	addEdgeLazy(addedNamedEdge) ;
    }

    /** **************************************************** 
     * Does not inforce a unique parent for each node as addEdge
     */
    private void addEdgeLazy(NamedEdge addedNamedEdge) throws java.lang.Exception {

	NamedVertex calleeGhost = (NamedVertex)addedNamedEdge.getSink() ;
	String nameOfCallee = calleeGhost.getName() ;
	addedNamedEdge.setName(nameOfCallee);

	NamedVertex remember = hiddenTopVertex ;
	addedNamedEdge = (NamedEdge)super.addEdgeReturnActualEdge(addedNamedEdge) ;

	// BEWARE: addEdgeReturnActualEdge may return an edge
	// different from the passed one. The caller and calle may not
	// be the original one.  (Though if everything goes right
	// still equivalent.)

	NamedVertex callee = (NamedVertex)addedNamedEdge.getSink() ;
	NamedVertex caller = (NamedVertex)addedNamedEdge.getSource() ;

	// We don't consider thread creation link to maintain the set
	// of top events (in getChildrenRegistryForVertex). This is
	// necessary to correctly read trace sets.

	String  subkind = addedNamedEdge.getValueForProperty("subkind") ;

	boolean isAThreadCreate = false ;
	if (subkind!=null) {
	    callee.setPropertyToValue ("kind", subkind ) ;
	    if (subkind.equals("threadCreate")) 
		isAThreadCreate = true ;
	}
	if ((!isAThreadCreate) && (underConstructionFromTraceSet)) {
	    topChildrenMap                      .removeChild(callee) ;
	    getChildrenRegistryForVertex(caller).   addChild(callee) ;
	}

    } // EndMethod addEdgeLazy

    protected void addThreadTermination(String threadID, 
                                        String eventID ) throws Exception 
    {
        NamedVertex threadStartPoint 
            = topChildrenMap.getYoungestChildForThread(threadID) ; 
        if (threadStartPoint==null) return ;
        String startEntityName = threadStartPoint.getValueForProperty("entity") ;
        String terminationCallPoint
            = ChildrenRegistry.getNameOfCallPoint( startEntityName, "<death>" ) ;

        NamedVertex threadEndPoint
            = new NamedVertex(terminationCallPoint+"'"+
                              CallGraphBuilder.getThreadTag(threadID)+eventID) ;

        threadEndPoint.setPropertyToValue("eventOrder",eventID) ;
        threadEndPoint.setPropertyToValue("threadID"  ,threadID) ;
        threadEndPoint.setPropertyToValue("entity"    ,startEntityName );
        threadEndPoint.setPropertyToValue("method"    ,"<death>" );

	add(threadEndPoint) ;
        
	NamedEdge newNamedEdge = 
	    new NamedEdge(threadStartPoint,threadEndPoint) ;

	setKind( threadEndPoint,newNamedEdge,"threadEnd") ;

	addEdgeLazy(newNamedEdge) ;
       
    } // EndMethod addThreadTermination

    /** **************************************************** */
    public void addCallFromTo(NamedVertex caller,NamedVertex callee) 
	throws java.lang.Exception 
    {
	NamedEdge newNamedEdge = 
	    new NamedEdge(caller,callee) ;
	String kind ;
        boolean isCPlusPlusObjectInstantiation = 
            MiscUtils.GetShortEntityName
            ( callee.getValueForProperty("entity") ).equals
            ( callee.getValueForProperty("method") ) ;

        boolean isJavaObjectInstantiation = 
            callee.getValueForProperty("method").equals("<init>") ;

	if ( isCPlusPlusObjectInstantiation || isJavaObjectInstantiation )
	    kind = "create" ;
	else
	    kind = "call" ;

	setKind( callee,newNamedEdge,kind ) ;
	addEdgeLazy(newNamedEdge) ;

    } // EndMethod addCallFromTo


    /** **************************************************** */
    public void addThreadCreationFromTo(NamedVertex caller,NamedVertex callee) 
	throws java.lang.Exception 
    {
	NamedEdge newNamedEdge = 
	    new NamedEdge(caller,callee) ;
	setKind( callee,newNamedEdge,"threadCreate") ;
	addEdgeLazy(newNamedEdge) ;
    }


    /** **************************************************** */
    private void setKind(NamedVertex callee, NamedEdge edgeOnCallee, String kind) {
	callee      .setPropertyToValue ("kind"   ,kind ) ;
	edgeOnCallee.setPropertyToValue ("subkind",kind ) ;
	edgeOnCallee.setPropertyToValue ("kind"   ,"controlflow" ) ;
    }

    /** **************************************************** 
     * returns a {@link DirectedGraphWriter} that
     * produces a condensed form of the {@link CallGraph} in
     * dot format.
     * @see CallGraphToCondensedGraphTransformer
     */
    public DirectedGraphWithDict2DotOStream getFormattingWriter() {
	return new CallGraphDotWriter() ;
    }

    /** **************************************************** */
    public NamedVertex getEventForID(int anEventID) {
	return (NamedVertex)_mapEventIDToVertex.get(new Integer(anEventID)) ;
    }

    /** **************************************************** 
     * This method is a bit like the mathematical floor function.
     * It tries to get the most recent event before a given event ID.
     * (Event IDs are a bit like logical time stamps ordering
     * observations.) If there exists an event for <code>anEventID</code>
     * then that very event is returned. If no event exists that matches
     * the passed ID, then the most recent event that occurred before that
     * ID is searched for.
     */
    public NamedVertex getMostRecentEventBeforeID(int anEventID) {
	SortedMap aux = _mapEventIDToVertex.headMap(new Integer(anEventID+1)) ;

	if (aux.size()==0) 
	    return null ;
	else 
	    return (NamedVertex)aux.get(aux.lastKey()) ;

    } // EndMethod getNearestOlderEventForID

    /** **************************************************** */
    public List getEventSBetweenEventID(int upperEventID, int lowerEventID ) {

	if (lowerEventID!=-1) // means infinum
	    if (lowerEventID<upperEventID) return new Vector() ;

	SortedMap aux ;

	if (lowerEventID==-1) // mean infinum
	    aux = _mapEventIDToVertex
		.tailMap(new Integer(upperEventID));
	else
	    aux = _mapEventIDToVertex
		.subMap(new Integer(upperEventID), new Integer(lowerEventID));

	return new Vector(aux.values()) ;

    } // EndMethod removeAfterEventID

    /** *********************************************************
     * Returns an Iterator that travels through the nodes of the graph
     * according to their event ID.
     */
    public Iterator getVerticesIteratorAccordingToEventID() {
	return _mapEventIDToVertex.values().iterator() ;
    }

    public Iterator getVerticesIterator() {
	return _mapEventIDToVertex.values().iterator() ;
    }


    public void changeParentOfNode(NamedVertex node,
				   NamedVertex newParent )
    	throws java.lang.Exception 
    {
	// Removing old parent link
	removeEdge( getParentEdge(node) ) ;

	// Creating a new link, while conserving thread creation information
	boolean    newParentIsThreadCreation =
	    newParent.getPropBag().hasPropertyToValue("kind","threadCreate") ;
	boolean    nodeIsSoucreCreation =
	    node   .getPropBag().hasPropertyToValue("kind","threadCreate") ;
	if ( newParentIsThreadCreation || 
	     nodeIsSoucreCreation )

	    addThreadCreationFromTo(newParent,node) ;
	else
	    addCallFromTo(newParent,node) ;
    } // EndMethod giveNodeNewParent

    public void moveOutGoingLinksFromNodeToNode(NamedVertex source,
						NamedVertex target )
	throws java.lang.Exception 
    {
	Collection outgoingS = getOutgoingAdjacentVertices(source) ;
	Iterator   iterOut   = outgoingS.iterator() ;
	boolean    sourceIsThreadCreation =
	    source.getPropBag().hasPropertyToValue("kind","threadCreate") ;

	while(iterOut.hasNext()) {

	    NamedVertex currentOutevent = (NamedVertex)iterOut.next() ;
	    boolean  currentOuteventIsThreadCreation =
		currentOutevent.getPropBag().hasPropertyToValue("kind","threadCreate") ;
	    if ( sourceIsThreadCreation || 
		 currentOuteventIsThreadCreation )

		addThreadCreationFromTo(target,currentOutevent) ;
	    else
		addCallFromTo(target,currentOutevent) ;

	} // EndLoop through outgoing verices

	// REMOVING ALL OUTGOING EDGES FROM SOURCE

	outgoingS = new Vector(getOutgoingEdges(source)) ;
	for (Iterator i = outgoingS.iterator() ; i.hasNext() ; ) {
	    NamedEdge currentEdge = (NamedEdge)i.next() ;
	    removeEdge(currentEdge) ;
	} // EndFor
	
    } // EndMethod moveOutGoingLinksFromNodeToNode

    /** **************************************************** 
     * removes a vertex from the graph, but connect each incoming neighbor
     * to each outgoing one.
     */
    public void abstractAway(NamedVertex abstractedEvent) throws java.lang.Exception {

	if (!this.doesContainVertex(abstractedEvent)) return ;
	Collection incomingS = getIncomingAdjacentVertices(abstractedEvent) ;
	if (incomingS.size()>1)
	    throw new RuntimeException
		("INTERNAL ERROR: A call graph contains an event "  +
		 "with more that one parent ! "+abstractedEvent+" " +
		 abstractedEvent.getPropBag() );
	if (incomingS.size()==1) {
	    NamedVertex thePredecessor = (NamedVertex)incomingS.iterator().next() ;

	    moveOutGoingLinksFromNodeToNode(abstractedEvent,thePredecessor) ;

	} // There is a predecessor

	remove(abstractedEvent) ;

    } // EndMethod abstractAway


    class TagForNodes implements Comparable {
	boolean nodeIsStart ;
	Integer eventID ;
	public TagForNodes(String anEventID, boolean isItStart) {
	    this.eventID     = Integer.valueOf(anEventID) ;
	    this.nodeIsStart = isItStart ;
	} // EndConstructor TagForNodes
	public int compareTo(Object o) {
	    TagForNodes autre = (TagForNodes)o ;
	    return this.eventID.compareTo(autre.eventID) ;
	} // EndMethod compareTo
    } // EndClass TagForNodes

    /**
     * createTagsMap is used by {@link rebind} and {@link
     * leapOverNodes}. It creates a map that sort the set of start
     * nodes and of end nodes according to their time ID (or event
     * ID). In order to distinguish between end and start nodes, we
     * use a subclass of String that contains a boolean tag to tell us
     * if this is start or end.
     */
    private SortedMap createTagsMap(Set startSet, Set endSet) {

	SortedMap eventIDToNodeMap = new TreeMap() ;

	for(Iterator i = startSet.iterator() ; i.hasNext() ; ) {
	    NamedVertex currentNode = (NamedVertex)i.next() ;
	    String currentEventID   = currentNode.getValueForProperty("eventOrder") ;
	    eventIDToNodeMap.put(new TagForNodes(currentEventID,true) , currentNode) ;
	} // EndFor on startSet

	for(Iterator i = endSet.iterator() ; i.hasNext() ; ) {
	    NamedVertex currentNode = (NamedVertex)i.next() ;
	    String currentEventID   = currentNode.getValueForProperty("eventOrder") ;
	    eventIDToNodeMap.put(new TagForNodes(currentEventID,false) , currentNode) ;
	} // EndFor on endSet

	return eventIDToNodeMap ;
	
    } // EndMethod createTagsMap

    /**
     * Use a tag map created by {@link createTagsMap} and move edges
     * of the end nodes to the start node. The boolean switch
     * determines which sets of edges are moves: either outgoing (this
     * is used by {@link leapOverNodes}) or incoming (used by {@link
     * rebind}).
     *
     * <p><b>Date:</b> Fri Aug 24 2007 17:44:58
     * @author Francois Taiani
     */
    private void visitTagsAndMoveEdgesOutgoingOrNot(SortedMap eventIDToNodeMap,
						    boolean   moveOutgoing )
	throws Exception
    {
	NamedVertex lastEncounteredStartNode = null ;
	for(Iterator i = eventIDToNodeMap.keySet().iterator() ; i.hasNext() ; ) {

	    TagForNodes currentTag  = (TagForNodes)i.next() ;
	    NamedVertex currentNode = (NamedVertex)eventIDToNodeMap.get(currentTag);
//  	    System.out.println("currentNode: "+currentNode) ;
	    if (currentTag.nodeIsStart) {
		lastEncounteredStartNode = currentNode ;
	    } else { // LEAP OVER
// 		System.out.println("Leap from: "+currentNode+" to: "+lastEncounteredStartNode) ;
		if (lastEncounteredStartNode!=null) {
		    if (moveOutgoing) 
			moveOutGoingLinksFromNodeToNode
			    (currentNode,lastEncounteredStartNode) ;
		    else {
			changeParentOfNode
			    (currentNode,lastEncounteredStartNode) ;
		    }
		} // EndIf (lastEncounteredStartNode!=null)
 
	    } // EndElse this is an end node. We have to make a "leap over"

	} // EndFor Loop on start and end nodes in order of their event IDs
    } // EndMethod visitTagsAndMoveEdgesOutgoingOrNot

    /** ****************************************************
     * Move links based on the succesion of invocations on the time line
     * for certains nodes. This is a quite weird feature whose need
     * arose because of the particular implementation of the thread
     * library in the Linux Kernel 2.4. In Linux 2.4 The "invisible" manager
     * thread of the pthread lib interacts with the rest of the
     * application in a "server" fashion through a pipe. 
     * All invocations to "clone" are realized through the same
     * thread (the thread manager), and no information is explicitly
     * available in the trace about the threads on behalf of
     * whom a thread was created.<p>
     * To fix that {@link #leapOverNodes(Set startOfLeap, Set enOfLeap)}
     * goes the whole graph through, and each time it finds
     * a node from <code>enOfLeap</code> it looks for the nearest
     * node available in <code>startOfLeap</code> that occured before the node
     * of <code>enOfLeap</code> and move all children form the 
     * node of <code>enOfLeap</code> to the node of <code>startOfLeap</code>.
     * (the link relations "leap" from the node of <code>startOfLeap</code> to
     * the children of the node <code>enOfLeap</code>, over that second node.)<p>
     * On the Linux the following action should solve our pb
     * <pre>
     * this.leapOverNodes( this.getNodeSWithRegExp("::pthread_create*"),
     *                     this.getNodeSWithRegExp("::clone'*" ) ) ;
     * </pre>
     * In conclusion: a real weird feature.
     */
    public void leapOverNodes(Set startOfLeap, Set endOfLeap) 
	throws java.lang.Exception 
    {

	// We first sort the set of start nodes and of end nodes
	// according to their time ID (or event ID). For that
	// we use a unique map. In order to distinguish
	// between end and start nodes, we use a subclass
	// of String that contains a boolean tag to
	// tell us if this is start or end.

	SortedMap eventIDToNodeMap = createTagsMap(startOfLeap, endOfLeap) ;
	visitTagsAndMoveEdgesOutgoingOrNot(eventIDToNodeMap, true) ;

    } // EndMethod leapOverNodes

    
    /** ****************************************************
      * rebind is similar to {@link leapover} in that it moves edges between 2
     * sets of nodes according to timestamps. The difference is that
     * rebind work directly the children being moves,
     * i.e. rebind("A'**","B'**") change the parents of all node in
     * "B'**" to be the closet preceding node in "A'**". In addition
     * rebind fails if there is any ambiguity in the children parent
     * mapping, i.e. if several children would end up with the same
     * new parent. In this case the underlying graph stays unchanged.
     *
     * <p><b>Date:</b> Fri Aug 24 2007 17:30:26
     * @author Francois Taiani
     */
    public void rebind(Set newParents, Set newChildren) 
	throws java.lang.Exception 
    {
	// same as leapOverNodes

	SortedMap eventIDToNodeMap = createTagsMap(newParents,newChildren) ;

	// First we check that no mapping is ambiguous, i.e. that each
	// children is allocated to one and only one parent.
	
	TagForNodes lastEncounteredStartTag = null ;
	for(Iterator i = eventIDToNodeMap.keySet().iterator() ; i.hasNext() ; ) {
	    TagForNodes currentTag  = (TagForNodes)i.next() ;
	    if (currentTag.nodeIsStart) {
		lastEncounteredStartTag = currentTag ;
	    }
	    else {
		if (lastEncounteredStartTag==null) {
		    NamedVertex currentNode = (NamedVertex)eventIDToNodeMap.get(currentTag);
		    throw new AmbiguousRebindException
			("Failed rebind operation: "+currentNode+" has no parent");
		} // EndIf lastEncounteredStartTag==null
		lastEncounteredStartTag = null ;
	    } // EndIfElse currentNode.nodeIsStart
	} // EndFor

	// Here we work on incoming edges, so we change the parents
	visitTagsAndMoveEdgesOutgoingOrNot(eventIDToNodeMap, false) ;

    } // EndClass rebind

    private SortedMap createSortedNodeMap(Set setOfNodes) {

	SortedMap result = new TreeMap() ;

	for(Iterator i = setOfNodes.iterator() ; i.hasNext() ; ) {
	    NamedVertex currentNode = (NamedVertex)i.next() ;
	    Integer currentEventID   = Integer.valueOf(currentNode.getValueForProperty("eventOrder")) ;
	    result.put(currentEventID , currentNode) ;
	} // EndFor on setOfNodes
	
	return result ;

    } // EndMethod createSortedNodeMap

    /** ****************************************************
     *
     * fuse works like {@link rebind}, except it pairs children
     * and parents according to the order in which they appear (first
     * children with first parent, second childre with second parent,
     * etc.). This approach allows fuse to work on interleaved
     * sequence of parents and children. It returns an exception if
     * the number of parents and children are not the same, or if a
     * parent was observed after the child that should be allocated to
     * it.
     *
     * <p><b>Date:</b> Fri Aug 24 2007 17:30:26
     * @author Francois Taiani
     */
    public void fuse(Set newParents, Set newChildren) 
	throws java.lang.Exception 
    {
	// same as leapOverNodes

	SortedMap sortedParents  = createSortedNodeMap(newParents ) ;
	SortedMap sortedChildren = createSortedNodeMap(newChildren) ;
	

	// First we check that no mapping is ambiguous, i.e. that
	// there is the same number of parents and children, and that
	// no parent is younger that its allocated child.
	
	if (sortedChildren.size()!=sortedParents.size())
	    throw new AmbiguousRebindException("There are not the same number of children ("+
					       sortedChildren.size()+") as parents ("+
					       sortedParents.size()+").");

	Iterator parentCursor = sortedParents .keySet().iterator() ;
	Iterator childCursor  = sortedChildren.keySet().iterator() ;

	while(parentCursor.hasNext()) {
	    if (((Integer)parentCursor.next()).compareTo((Integer)childCursor.next())>0)
		throw new AmbiguousRebindException("Some parent is younger than is potential child.");
	} // EndWhile

	parentCursor = sortedParents .values().iterator() ;
	childCursor  = sortedChildren.values().iterator() ;

	// Here we work on incoming edges, so we change the parents

	while(parentCursor.hasNext()) {

	    changeParentOfNode ((NamedVertex)childCursor.next(),
				(NamedVertex)parentCursor.next()) ;

	} // EndWhile

    } // EndClass fuse

    /** ****************************************************
     * returns true is the passed named vertex is a call on
     * the same class as the calling one. (Internal call.)
     */
    public boolean eventIsASelfCall(NamedVertex possibleSelf) {
	NamedVertex parent = getParent(possibleSelf) ;
	if (parent!=null) {
	    String entityForPosSelf = possibleSelf.getValueForProperty("entity") ;
	    String entityFotParent  =       parent.getValueForProperty("entity") ;
	    if ( entityForPosSelf.equals(entityFotParent) ) {
		// Self has same entity has its parent event,
		// so that's certainly a recursive call, *except*
		// when the entity attribute is empty, which
		// correspond to a plain function outside any
		// class. (Like a C function for instance.
		if (entityForPosSelf.length()>0)
		    return true ;
		// So now we now that's a C function. For the call
		// event to be recursive, the "method" attributes must be 
		// equal.
		return (  possibleSelf.getValueForProperty("method").equals
			  (     parent.getValueForProperty("method") ) ) ;
	    } // EndIf same entity attribute (but possibly empty)
	} // EndIf is not a root event
	return false ;
    } // EndMethod eventIsASelfCall

    /** ****************************************************
     * removes call from within a class to the same class. Calls
     * are abstracted away. (See {@link #abstractAway(NamedVertex)}).
     * To have the action apply on all calls of a given class, you
     * can use a regexp of the form '<NameOfTheClass>::*'. (see 
     * {@link DirectedGraphWithDictImpl#getNodeSWithRegExp(String)}.)
     */
    public void removeSelfCallOnNodeMatchingRegexp(String regularExpressionName)
	throws java.lang.Exception 
    {
	SortedSet matchingNodes = getNodeSWithRegExp(regularExpressionName) ;

	// LOOP ON ALL MATCHING NODES
	for(Iterator i = matchingNodes.iterator() ; i.hasNext() ; ) {
	    NamedVertex currentEvent = (NamedVertex)i.next() ;
	    if (eventIsASelfCall(currentEvent)) abstractAway(currentEvent) ;
	} // EndFor all matching nodes

    } // EndMethod removeSelfCallOnNodeMatchingRegexp

    public String getHumanFriendlyName(NamedVertex toBeNamed, 
				       boolean     removeNamespace) {
	String globalOrder ;
	Integer symbolicID = getSymbolicIDForInvocationVertex(toBeNamed);
	if (symbolicID!=null) { 
	    globalOrder = symbolicID.toString();
	} else {
	    globalOrder = toBeNamed.getValueForProperty("eventOrder") ;
	}
	String threadIDs     = toBeNamed.getValueForProperty("threadID"  ) ;
	String entityName    = toBeNamed.getValueForProperty("entity") ;
	if (removeNamespace) 
            entityName = MiscUtils.GetShortEntityName(entityName) ;
	String invokedMethod = 
	    entityName + "::" + toBeNamed.getValueForProperty("method") ;

	return ConstructInvocationString(globalOrder,threadIDs,invokedMethod) ;

    } // EndMethod getHumanFriendlyName

    /** ****************************************************
     * Constructs a string representing an invocation from several
     * substrings representing respectively the global order of the
     * invocation, the thread ids (possibly several) of the acting
     * thread, and the invoked method. This method is used by {@link
     * #getHumanFriendlyName(NamedVertex,boolean)} for raw call
     * graphs, and by {@link CallLinkRenderer} for condensed call
     * graphs.
     */
    public static String ConstructInvocationString(String globalOrder,
						   String threadIDs,
						   String invokedMethod )
    {
	StringBuffer result = new StringBuffer() ;
	if (globalOrder!=null)
	    if (globalOrder.length()>0)    result.append("("+globalOrder+") ") ;
	if (threadIDs!=null)
	    if (threadIDs  .length()>0)    result.append("t"+threadIDs) ;
	if ((invokedMethod!=null) && (threadIDs!=null))
	    if ((invokedMethod.length()>0) && 
		(threadIDs .length()>0) )  result.append(": ") ;
	if (invokedMethod!=null)
	    if (invokedMethod .length()>0)    result.append( invokedMethod ) ;
	return result.toString() ;

    } // EndMethod ConstructInvocationString

    protected Map globalIDToSymbolicIDMap = null ;

    public Integer getSymbolicIDForInvocationVertex(NamedVertex anInvocation) {
	if (globalIDToSymbolicIDMap==null) return null ;
	return (Integer)globalIDToSymbolicIDMap.get(anInvocation);
    } // End getSymbolicIDForInvocationVertex

    /**
     * When about to be visualised, this method creates symbolic IDs
     * for all invocation nodes in the call graph.
     */
    public void prepareForVisualisation() {
	globalIDToSymbolicIDMap = new TreeMap() ;
	Iterator iterOverVertices = getVerticesIteratorAccordingToEventID();
	int count = 0;
	while(iterOverVertices.hasNext()) {
	    NamedVertex node = (NamedVertex)iterOverVertices.next();
	    globalIDToSymbolicIDMap.put(node,new Integer(count));
	    count++;
	} // EndWhile
    } // EndMethod prepareForVisualisation

    /**
     * This releases the invocation ID maps generated by {@link
     * prepareForVisualisation()}
     */
    public void endOfVisualisation() {
	globalIDToSymbolicIDMap = null ;
    } // EndMethod endOfVisualisation

} // EndClass CallGraph

class AmbiguousRebindException extends salvo.jesus.graph.GraphException {
    public AmbiguousRebindException(String msg) {
	super(msg);
    } // EndConstructor
} // EndClass AmbiguousRebindException
