/*
 * Copyright 2003 LAAS-CNRS
 *
 *   LAAS-CNRS
 *   7, avenue du Colonel Roche
 *   F-31077 Toulouse Cedex 4 / France
 *
 * Author: Francois Taiani
 *
 * This file is part of OPSBrowser.
 * 
 * OPSBrowser is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * OPSBrowser is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with OPSBrowser; if not, write to the Free Software Foundation,
 * Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 * 
 */
/**
 * CallGraphToCondensedGraphTransformer.java
 * 
 * Time-stamp: <2003-03-11 10:48:42 ftaiani>
 * @author Francois Taiani   <taiani@laas.fr>
 * CVS Info: $Id: CallGraphToCondensedGraphTransformer.java,v 1.6 2007/09/18 18:23:42 ftaiani_normal Exp $
 * 
 */

package laas.friends.cosmopen.graphAnalysis ;

import java.util.* ;
import laas.friends.cosmopen.utils.*  ;

/** ***********************************************
 * @author Francois Taiani   <taiani@laas.fr>
 * @version $Name: release-1-3-rc1 $ $Revision: 1.6 $
 * @since Version 0.1.1
 */
public class CallGraphToCondensedGraphTransformer {

    protected CondensedGraph result   ;
    private CallGraph _transformee  ;
    private NamedVertex virtualFirstInvocation ;

    protected Map     invocation2GlobalSymbolicID ;
    private boolean olderExist ;
    private boolean youngerExist ;

    static protected boolean explicitRange = false ;
    public static void setExplicitRange(boolean newValue) { 
	explicitRange = newValue ; 
    } 

    /** ****************************************************
     * Transforms a {@link CallGraph} into a class interaction graph.
     * This method is a bit technical. A callgraph is a union of trees
     * (AKA a forest).  Transforming it into an interaction graph
     * includes a sort of aggregating dualization: the nodes of the
     * callgraph become links in the resulting interaction graph. The
     * nodes of this new graph are the entity on which the calls
     * represented by the nodes of the callgraph are made.
     * @param transformee the {@link CallGraph} to be transformed
     * @return            the interaction graph.
     */
    public DirectedGraphWithDict fromCallGraphToGraph(CallGraph transformee) throws Exception {

	result = new CondensedGraph() ;
	_transformee = transformee ;

	// Used as a trick to handle parentless invocation (no previous invocation), and
	// in particular the "precedent" tag in the corresponding result graph.

	virtualFirstInvocation = new NamedVertex("%% VIRTUAL START %%") ;
	virtualFirstInvocation.setPropertyToValue("entity","") ;
	virtualFirstInvocation.setPropertyToValue("method","") ;

	invocation2GlobalSymbolicID = new HashMap() ;

// 	Iterator iterOverVertices = transformee.getVerticesIterator() ;
	Iterator iterOverVertices = transformee.getVerticesIteratorAccordingToEventID() ;

	// LOOP ON ALL NODES ("INVOCATIONS") OF THE INPUT CALLGRAPH 

	//     Each invocation is transformed into an interaction link in
	//     the condensed graph. Vertices in the condensed graph
	//     are created on the fly. Only "local" properties are set
	//     in this loop. "Navigation" properties that explicitize
	//     the causal succession of calls on classes are not in
	//     this while loop but in the following while loop. Some
	//     information is gathered here by the structured of the
	//     CondensedVertex class that is used to represent the
	//     vertices of the condensed graph (instead of plain
	//     NamedVertex). This information is used afterward to set
	//     the "Navigation properties.

	int globalInvocationCount = 0 ;

	while(iterOverVertices.hasNext()) {
	    NamedVertex currentInvocation  = (NamedVertex)iterOverVertices.next() ;
	    transformInvocationIntoInteractionInCondensedGraph(currentInvocation) ;

	    // Root invocations won't appear in the final condensed
	    // graph, so they should not get any global symbolic ID.

	    if (_transformee.getParentEdge(currentInvocation)!=null) {
		globalInvocationCount++ ;
		invocation2GlobalSymbolicID.put( currentInvocation, 
						 new Integer(globalInvocationCount) ) ;
	    }

	} // EndWhile ( transformee.getVerticesIterator().hasNext() )

	// END LOOP ON ALL NODES OF THE INPUT CALLGRAPH

	// Very inneficient but robust :(

	iterOverVertices = result.getVerticesIterator() ;

	// LOOP ON ALL "INTERACTION" LINKS OF THE RESULTING CONDENSED GRAPH
	//
	//    Here we set the "Navigation" properties that explicitize
	//    the causal succession of calls on classes.

	while(iterOverVertices.hasNext()) {
	    CondensedVertex currentVertex = (CondensedVertex)iterOverVertices.next() ;
	    currentVertex.labelEdges(this) ;
	} // EndWhile

	// SOME CLEANING UP, AND WE'RE FINISHED

	result.remove(retrieveNodeForInvocation(virtualFirstInvocation)) ;

	// Improving garbage collection (very bad style !)
	CondensedGraph tempResult = result ;

	_transformee = null ;
	result = null ;

	return tempResult ;

    } // EndMethod fromCallGraphToGraph

    private void checkAgeOfNodeAgainstBrotherS(NamedVertex child,NamedVertex parent) {

	olderExist  = false ;
	youngerExist = false ;
	Iterator brotherS = _transformee.getOutgoingAdjacentVertices(parent).iterator() ;
	int idCandidate = Integer.parseInt(child.getValueForProperty("eventOrder")) ;

	while((brotherS.hasNext()) && !(olderExist&&youngerExist)) {
	    NamedVertex currentBrother = (NamedVertex)brotherS.next() ;
	    int currentID = Integer.parseInt(currentBrother.getValueForProperty("eventOrder")) ;
	    if (idCandidate < currentID) olderExist = true ;
	    if (idCandidate > currentID) youngerExist= true ;
	}
	// No brother was found younger that candidate, so that's the youngest
	return ;
	
    } // End checkAgeOfNodeAgainstBrotherS

    /** ****************************************************
     * Creates a new interaction link in the condensed graph to
     * be created. This new interaction link correspond to the
     * passed <code>callInvocationToBeTransformed</code>. Only
     * the "local" properties of the new interaction link
     * are set: "isRoot", "kind", "method", "isReturn", "isStart",
     * "isSameObject", "newThread", "isLeaf".
     */
    private void transformInvocationIntoInteractionInCondensedGraph
	( NamedVertex callInvocationToBeTransformed ) throws Exception
    {

	NamedVertex parentInvocation = 
	    getParentWithinTransformee(callInvocationToBeTransformed)  ;

	if (parentInvocation==null) {
	    if (_transformee.getOutgoingEdges(callInvocationToBeTransformed).isEmpty())
		// This invocation won't appear at all on the final graph,
		// since it has neither predecessor nor successors
		// so we don't take it into account to avoind border effects
		// on the sequencing of calls on a given entity.
		return ;
	    parentInvocation = virtualFirstInvocation ;
	}
	CondensedVertex caller = retrieveNodeForInvocation ( parentInvocation ) ;
	CondensedVertex callee = retrieveNodeForInvocation ( callInvocationToBeTransformed  ) ;
	
	if (parentInvocation==virtualFirstInvocation)
	    callee.setPropertyToValue("isRoot","y");

	// START SETTING UP THE NEW LINK
	    
	NamedEdge invocationAsLink = new NamedEdge(caller,callee) ;

	String threadID = callInvocationToBeTransformed.getValueForProperty("threadID"  ) ;
	String method   = callInvocationToBeTransformed.getValueForProperty("method") ;

	// Those are the standard properties. They may be changed however
	// by the "if"-block that handles the presence of a parent invocation,
	// in particular if the current invocation represent a thread creation.

	invocationAsLink.setPropertyToValue( "kind"  , callInvocationToBeTransformed
                                                  .getValueForProperty("kind")) ;
	invocationAsLink.setPropertyToValue( "method", method   ) ;

	String threadIndicatorInName ;

	if (threadID!=null) 
	    threadIndicatorInName =threadID+":"; 
	else  threadIndicatorInName = "" ;

	// SETTING PROPERTIES THAT ARE BASED ON COMPARISON WITH PARENT INVOCATION
	if (parentInvocation!=virtualFirstInvocation) {

	    // HANDLING ISsTART AND ISrETURN PROPERTIES
	    checkAgeOfNodeAgainstBrotherS(callInvocationToBeTransformed,parentInvocation);
	    if (!olderExist  ) invocationAsLink.setPropertyToValue( "isReturn","y" ) ;
	    if (!youngerExist) invocationAsLink.setPropertyToValue( "isStart" ,"y" ) ;

	    // RECOGNIZING SELF INVOCATION OF OBJECTS ("this->method()")
	    String previousThis = parentInvocation  .getValueForProperty("this") ;
	    String current_This = callInvocationToBeTransformed .getValueForProperty("this") ;
	    if (previousThis!=null) 
		if (previousThis.length()>0)
		    if (previousThis.equals(current_This))
			invocationAsLink.setPropertyToValue( "isSameObject","y" ) ;

	    // HANDLING THREAD CREATION
	    String parentThread = parentInvocation  .getValueForProperty("threadID") ;
	    if (parentThread!=null) {
		if (!parentThread.equals(threadID)) {
		    invocationAsLink.setPropertyToValue( "kind"      , "threadCreate" );
		    invocationAsLink.setPropertyToValue( "newThread" , threadID );
		    threadIndicatorInName = parentThread+"N"+threadID+":" ;
		    // This last line means we want the name of the thread creation link 
		    // to include both the parent and the child thread. (For
		    // link unicity constraints).
		    threadID = parentThread ;
		}
	    } // The parent invocation has a thread ID

	} // EndIf linkToParentInvocation!=null

	invocationAsLink.setName  ( threadIndicatorInName+method ) ;
	invocationAsLink.setPropertyToValue( "thread"    , threadID ) ;

	if (_transformee.getOutDegree(callInvocationToBeTransformed)==0) // No children invocation
	    invocationAsLink.setPropertyToValue( "isLeaf","y" ) ;

	// System.out.println("New invocation link : "+invocationAsLink.getPropBag() ) ;

	invocationAsLink = result.addReturnEdgeAndStrictOrNot(invocationAsLink,false) ;

	// System.out.println(invocationAsLink.getName()+":"+
	//                    invocationAsLink+"="+invocationAsLink.hashCode()) ;
	// System.out.println("caller:" + caller + " " + caller.getPropBag() ) ;
	// System.out.println("callee:" + callee + " " + callee.getPropBag() ) ;
	// System.out.println("New Link:" + invocationAsLink + " " 
	//                                + invocationAsLink.getPropBag() ) ;

	caller.notifyOutgoingCall( invocationAsLink , callInvocationToBeTransformed ) ;
	callee.notifyIncomingCall( invocationAsLink , callInvocationToBeTransformed ) ;

    } // EndMethod transformInvocationIntoInteractionInCondensedGraph

    private CondensedVertex retrieveNodeForInvocation(NamedVertex anInvocation) {

	String entity = anInvocation.getValueForProperty("entity") ;
	String method = anInvocation.getValueForProperty("method") ;
	String nameOfNode ;
	if (entity==null) throw new RuntimeException
			      (anInvocation+" "+anInvocation.getPropBag()+" "+
			       "has no entity !") ;
	if (entity.length()==0)  nameOfNode = method ;
	else                     nameOfNode = entity ;

	CondensedVertex resultingVertex = 
	    (CondensedVertex)result.provideNamedVertexWithName(nameOfNode) ;

	//	System.out.println("CREATING CONDENSED VERTEX:"+resultingVertex+" for "+anInvocation);

	if (entity.length()==0)  
	    resultingVertex.setPropertyToValue("isFunction","y") ;

	return resultingVertex ;
	
    } // EndMethod retrieveNodeForInvocation

    /** ****************************************************
     * returns the parent node is any of the passed {@link NamedVertex}.
     * The parent is searched within the {@link CallGraph} that is 
     * being transformed.
     * @see #fromCallGraphToGraph(CallGraph transformee)
     */
    public NamedVertex getParentWithinTransformee(NamedVertex lostChild) {

	NamedEdge   linkToParentInvocation = _transformee.getParentEdge(lostChild);
	
	if (linkToParentInvocation==null) 
		return null ;

	return (NamedVertex)linkToParentInvocation.getSource() ;

    } // EndMethod getParentWithinTransformee


} // EndClass CallGraphToCondensedGraphTransformer

/** ****************************************************
 * The type of vertices that is used in {@link CondensedGraph}.
 * This kind of vertex maintain various structures in order
 * to provide symbolic representation of the causality of
 * calls on a class / object, in order to help navigate
 * an interaction graph.<p>
 *
 * In an interaction graph, a link between 2 classes may
 * represent several different invocations, each of these
 * invocation corresponding to a particular invocation of
 * the sink class from the source class.<p>
 *
 * A {@link CondensedVertex} maintain tables about
 * all invocations that arrive on it, and all invocations
 * that go out of it. It also remembers which invocations a given
 * "interaction link" actually stands for, since the same
 * "interaction link" may represent several invocations between
 * the same 2 classes.<p>
 *
 * All this information is used to label the *interaction* links
 * (which is what the user will see utimately) with symbolic ranges
 * (or number in case the range contains a unique symbol) that reflect
 * the succession of *invocation* (calls) on classes.
 */
class CondensedVertex extends NamedVertex {

    public CondensedVertex () {
    } // EndConstructor NamedVertex

    public CondensedVertex (String nameForNewNamedVertex) {
	super(nameForNewNamedVertex) ;
    } // EndConstructor NamedVertex(String nameForNewNamedVertex)

    /** 
     * Maps outgoing invocations (NamedVertex in the original graph) 
     * to interaction links (NamedEdge in the resulting condensed
     * graph.)
     */
    private SortedMap outgoingInvocation2outgoingEdges = new TreeMap() ;

    /** 
     * Maps incoming invocation (NamedVertex in the original graph) to
     * interaction links (NamedEdge in the resulting condensed graph).
     */
    private SortedMap incomingInvocation2incomingEdges = new TreeMap() ;

    private Map outgoingInvocation2symbol ;
    private Map incomingInvocation2symbol ;

    public void notifyIncomingCall(NamedEdge   incomingInteractionLink, 
				   NamedVertex incomingInvocationsNode ) 
    {
	incomingInvocation2incomingEdges.put( incomingInvocationsNode,
					      incomingInteractionLink ) ;
    } // EndMethod notifyIncomingCall

    public void notifyOutgoingCall(NamedEdge   outgoingInteractionLink, 
				   NamedVertex outgoingInvocationsNode ) 
    {
// 	System.out.println("   XXXXXXXXX New outgoing invocation: "+outgoingInvocationsNode+" "+
// 			   outgoingInteractionLink);
	outgoingInvocation2outgoingEdges.put( outgoingInvocationsNode,
					      outgoingInteractionLink ) ;
    } // EndMethod notifyOutgoingCall

    public void labelEdges(CallGraphToCondensedGraphTransformer theTransformer) {
	
	//	System.out.println("labelEdges on "+this);

	boolean manyOutgoingEdges = outgoingInvocation2outgoingEdges.size()>1 ;
	boolean manyIncomingEdges = incomingInvocation2incomingEdges.size()>1 ;

	if (manyOutgoingEdges) handleOutgoingEdges() ;
	if (manyIncomingEdges) {
	    handleIncomingEdges() ;
	    linkOutEdges2InEdges(theTransformer) ;
	}

	setGlobalSymbolicOrderID(theTransformer) ;

	// Garbage collection
	outgoingInvocation2outgoingEdges  = null ;
	incomingInvocation2incomingEdges  = null ;

    } // EndMethod labelEdges

    
    private Map generateInvocation2ParentSymbolMap
	( Set outgoingInvocationS,
	  Map incomingInvocation2LocalSymbolicID,
	  CallGraphToCondensedGraphTransformer transformer)
    {
	Iterator iterOnInvocations = outgoingInvocationS.iterator() ;
	Map invocation2ParentSymbol = new HashMap() ;
	while(iterOnInvocations.hasNext()) {
	    NamedVertex currentInvocation = (NamedVertex)iterOnInvocations.next() ;
	    NamedVertex currentParent     = 
		transformer.getParentWithinTransformee(currentInvocation) ;
	    if (currentParent==null) continue ;
	    Integer symbol 
		= (Integer)incomingInvocation2LocalSymbolicID.get(currentParent) ;
	    if (symbol!=null)
		invocation2ParentSymbol.put(currentInvocation,symbol) ;
	} // EndWhile

	return invocation2ParentSymbol ;

    } // EndMethod generateInvocation2ParentSymbolMap


    /** ****************************************************
     * This method is used to compute symbolic range representation
     * for interaction links (that are represented in the condensed
     * graph) from invocation "invocations" (the nodes of the original
     * {@link CallGraph}).
     * @see CallGraphToCondensedGraphTransformer
     * @see #handleOutgoingEdges()
     * @see #handleIncomingEdges()
     */
    private Map computeRangesForEachInteractionEdge(SortedMap invocation2Interaction,
						    Map       invocation2symbol )
    {
	Map interactionToRange = new HashMap() ;
	Iterator iterInvocationS = invocation2Interaction.keySet().iterator() ;

// 	System.out.println("NEW MAPPING\n\n") ;

	while(iterInvocationS.hasNext()) {
	    NamedVertex currentInvocation
		= (NamedVertex)iterInvocationS.next() ;
	    NamedEdge  currentInteraction 
		= (NamedEdge)invocation2Interaction.get(currentInvocation) ;
	    OrderRange currentRange = (OrderRange)interactionToRange.get(currentInteraction) ;
	    Integer callOrder = (Integer)invocation2symbol.get(currentInvocation) ;
	    if (callOrder==null) {
		//		System.out.println("Call order is null. continuing");
		continue ;
	    }

	    if (currentRange==null) {
		currentRange = createNewOrderRange(callOrder.intValue());
		interactionToRange.put(currentInteraction,currentRange) ;
	    } else
		currentRange.addToRange(callOrder.intValue()) ;

//   	    System.out.println(currentInteraction+" "+currentInteraction.getName() +" "+ callOrder + 
//  			       " => " + currentRange ) ;
// 	    System.out.println("Out Link:" + currentInteraction + " " + 
// 			       currentInteraction.getPropBag() ) ;

	} // EndWhile

	return interactionToRange ;

    } // EndMethod computeRangesForEachInteractionEdge

    private OrderRange createNewOrderRange(int startValue) {
	if (CallGraphToCondensedGraphTransformer.explicitRange) 
	     return new ExplicitOrderRange(startValue) ;
	else return new MinMaxOrderRange  (startValue) ;
    }

    /** ****************************************************
     * Labels the outgoing interaction links of the {@link CondensedVertex}.
     * The labeling involves a two step transformation. First the
     * invocation vertices of the original {@link CallGraph} that originate
     * from the class / object represented by the {@link CondensedVertex}
     * must be labeled with symbols according to their eventOrder attribute.
     * Then a range representation must be compute for each interaction
     * edge of the resulting condensed graph according to the symbols
     * allocated to the invocation nodes this interaction stands for.
     */
    private void handleOutgoingEdges() {

	outgoingInvocation2symbol = 
	    generateSymbolMapFromInvocationSet
	    (  outgoingInvocation2outgoingEdges.keySet() ) ;

	genericInteractionLinkLabeling
	    ( outgoingInvocation2outgoingEdges,
	      outgoingInvocation2symbol,
	      "callOrder",
	      OrderRange.intFormat,
	      null ) ;

    } // EndMethod handleOutgoingEdges


    /** ****************************************************
     * See {@link #handleOutgoingEdges()}.
     */
    private void handleIncomingEdges() {

	incomingInvocation2symbol = 
	    generateSymbolMapFromInvocationSet
	    (  incomingInvocation2incomingEdges.keySet() ) ;

	genericInteractionLinkLabeling
	    ( incomingInvocation2incomingEdges,
	      incomingInvocation2symbol,
	      "incomingOrder",
	      OrderRange.alphaFormat,
	      null ) ;

    } // EndMethod handleIncomingEdges

    /**
     * This method transforms individual IDs associated with
     * invocations in the original graph into strings representing
     * range to associate them with interaction link in the condensed
     * graph being constructed.
     *
     * This is need because an interaction link in the condensed graph
     * can stand for several invocation links of the original call
     * graph, so that any information (here a symbolic ID reflecting
     * some order on the sequence of incoming or outgoing invocations,
     * locally to one class, or globally on the graph) associated with
     * invocations must be condensed in some range to be associated
     * with interaction links. This method does that. The hashmap
     * invocation2InteractionLink specifies how invocations of the
     * original call graph are associated with interaction links of
     * the condensed graph in a n:1 fashion, and invocation2SymbolicID
     * specifies how invocations of the original call graph are
     * associated with symbolic IDs in a 1:1 fashion.
     *
     * symbolicID2InteractionLink can be used to return a map from the
     * allocated symbolicID to the interaction links if needed. This
     * happens when symbolicID2InteractionLink is non-nul, otherwise
     * it is just ignored.
     */
    private void genericInteractionLinkLabeling( SortedMap invocation2InteractionLink,
						 Map       invocation2SymbolicID,
						 String    propertyToBeSet,
						 int       rangeFormatID,
						 SortedMap symbolicID2InteractionLink)
    {
	Map interactionLinkToOrderRange = 
	    computeRangesForEachInteractionEdge
	    (    invocation2InteractionLink,
		 invocation2SymbolicID ) ;

	Iterator iterInteraction = interactionLinkToOrderRange.keySet().iterator() ;

	while(iterInteraction.hasNext()) {
	    NamedEdge  currentInteraction = (NamedEdge)iterInteraction.next() ;
	    OrderRange currentRange    = 
		(OrderRange)interactionLinkToOrderRange.get(currentInteraction) ;
	    String     orderAsString   = currentRange.toFormattedString(rangeFormatID) ;
	    currentInteraction.setPropertyToValue ( propertyToBeSet, orderAsString ) ;
	    if (symbolicID2InteractionLink!=null) {
		//		System.out.println("Generating Map: "+currentRange.min+" - "+currentInteraction);
		symbolicID2InteractionLink.put(new Integer(currentRange.min),
					       currentInteraction) ;
	    } // EndIf
	} // EndWhile

    } // EndMethod genericInteractionLinkLabeling
					       

    private void linkOutEdges2InEdges(CallGraphToCondensedGraphTransformer theTransformer)
    {
	Map outgoingInvocation2ParentSymbol = 
	    generateInvocation2ParentSymbolMap
	    ( outgoingInvocation2outgoingEdges.keySet(),
	      incomingInvocation2symbol,
	      theTransformer ) ;
	
	Map outgoingEdgeToParentOrderRange = 
	    computeRangesForEachInteractionEdge
	    ( outgoingInvocation2outgoingEdges,
	      outgoingInvocation2ParentSymbol ) ;

	Iterator iterOutgoing = outgoingEdgeToParentOrderRange.keySet().iterator() ;
	while(iterOutgoing.hasNext()) {
	    NamedEdge  currentOutGoing = (NamedEdge)iterOutgoing.next() ;
	    OrderRange currentRange    = 
		(OrderRange)outgoingEdgeToParentOrderRange.get(currentOutGoing) ;
 	    currentOutGoing.setPropertyToValue("precedent",currentRange.toAlphaString()) ;
	} // EndWhile

    } // EndMethod linkOutEdges2InEdges

    private void setGlobalSymbolicOrderID
	(CallGraphToCondensedGraphTransformer theTransformer)
    {

// 	System.out.println("START setGlobalSymbolicOrderID");
// 	System.out.println("outgoingInvocation2outgoingEdges: "+outgoingInvocation2outgoingEdges);
	
	genericInteractionLinkLabeling
	    ( outgoingInvocation2outgoingEdges,
	      theTransformer.invocation2GlobalSymbolicID,
	      "globalOrder",
	      OrderRange.intFormat,
	      theTransformer.result.symbolicID2InteractionLink) ;
// 	System.out.println("END setGlobalSymbolicOrderID");
	
    } // EndMethod setGlobalSymbolicOrderID

    /** *************************************************************
     * This method labels a set of invocation nodes with symbols based
     * on their eventID attribute.  In a condensed grpah, links
     * arriving on and leaving from {@link CondensedVertex}'s are
     * labelled with symbol that reflect their order of arrival /
     * leaving. Labelling the invocations of the original graph that
     * stay behind the link found in a coundensed graph is a first
     * step to that symbolic labeling.
     */
    protected static Map generateSymbolMapFromInvocationSet(Set invocationS) {
	Iterator iterOnInvocations = invocationS.iterator() ;
	SortedMap eventIDToInvocation = new TreeMap() ;
	while(iterOnInvocations.hasNext()) {
	    NamedVertex currentInvocation = (NamedVertex)iterOnInvocations.next() ;
	    Integer eventID 
		= Integer.decode(currentInvocation.getValueForProperty("eventOrder") ) ;
	    eventIDToInvocation.put(eventID,currentInvocation) ;
	} // EndWhile

	Map invocation2LocalSymbolicID = new HashMap() ;
	iterOnInvocations = eventIDToInvocation.values().iterator() ;
	int symbolicOrder = 0 ;
	while(iterOnInvocations.hasNext()) {
	    symbolicOrder++ ;
	    NamedVertex currentInvocation = (NamedVertex)iterOnInvocations.next() ;
	    invocation2LocalSymbolicID.put(currentInvocation, new Integer(symbolicOrder)) ;
	} // EndWhile
	return invocation2LocalSymbolicID ;

    } // EndMethod generateSymbolMapFromInvocationSet

} // EndClass CondensedVertex

/** ****************************************************************
 * The type of graph used in {@link
 * CallGraphToCondensedGraphTransformer} to contain the result of the
 * transformation. The difference to {@link DirectedGraphWithDictImpl}
 * is that this kind of graph uses {@link CondensedVertex} for its
 * nodes instead of plain {@link NamedVertex}.
 */
class CondensedGraph extends DirectedGraphWithDictImpl {

    protected SortedMap symbolicID2InteractionLink = new TreeMap();

    public java.util.Iterator getEdgeIterator() {
	if (symbolicID2InteractionLink==null) {
	    return super.getEdgeIterator();
	} else {
// 	    System.out.println("Here");
	    java.util.Iterator result = symbolicID2InteractionLink.values().iterator() ;
// 	    while(result.hasNext()) 
// 		System.out.println(result.next());
	    return symbolicID2InteractionLink.values().iterator() ;
	} // EndIfElse
    } // EndMethod getEdgeIterator


    public NamedVertex createNewNameVertexWithName(String aName) {
	return  new CondensedVertex(aName) ;
    } // EndMethod createNewNameVertexWithName

    public void prepareForVisualisation() {}

    public void endOfVisualisation() {}

    public boolean isEdgeOrderProeminant() { return true ; }

} // EndClass CondensedGraph

/**
 * abstract class that is used to represented ranges of numbers or
 * letters in a condensed graph. Such ranges are needed because a
 * single edge in a condensed graph may represent several invocations
 * of the original call graph.
 */
abstract class OrderRange {

    public OrderRange(int newInt) { min = newInt ; max = newInt ; }

    public static final int intFormat   = 0 ;
    public static final int alphaFormat = 1 ;

    public int min = 0 ;
    public int max = 0 ;

    public void addToRange(int newInt) {
	if ( newInt>max ) max = newInt ;
	if ( newInt<min ) min = newInt ;
    } // EndMethod addToRange

    public abstract String toString();
    public abstract String toAlphaString();
    public String toFormattedString(int formatID) {
	switch(formatID) {
	case intFormat   : return toString() ;
	case alphaFormat : return toAlphaString() ;
	default :
	    throw new RuntimeException("Unknow range format ID.") ;
	} // EndSwitch
    }

} // EndInterface OrderRange

class ExplicitOrderRange extends OrderRange {

    private SortedSet _content = new TreeSet() ;

    public ExplicitOrderRange(int newInt) { 
	super(newInt);
	_content.add(new Integer(newInt)) ; 
    } 

    public void addToRange(int newInt) {
	super.addToRange(newInt);
	_content.add(new Integer(newInt)) ; 
    } // EndMethod addToRange

    public String toString() {
	if (_content.isEmpty()) return "" ;
	Iterator i = _content.iterator() ;
	StringBuffer  result = new StringBuffer(((Integer)i.next()).toString()) ;
	while( i.hasNext() ) {
	    result
		.append(',') 
		.append(((Integer)i.next()).toString()) ;
	} // EndFor i

	return result.toString() ;
    } // EndMethod toString

    public String toAlphaString() {
	if (_content.isEmpty()) return "" ;
	Iterator i = _content.iterator() ;
	StringBuffer  result = new StringBuffer() ;
	result.append(MiscUtils.charCoding(((Integer)i.next()).intValue())) ;
	while( i.hasNext() ) {
	    result
		.append(",") 
		.append(MiscUtils.charCoding
			(((Integer)i.next()).intValue())) ;
	} // EndFor i

	return result.toString() ;

    } // EndMethod toAlphaString

} // EndClass ExplicitOrderRange

class MinMaxOrderRange extends OrderRange { 

    public MinMaxOrderRange(int newInt) { super(newInt); }

    public String toString() {
	StringBuffer  result = new StringBuffer(String.valueOf(min)) ;
	if (max!=min) result.append('-').append(max) ;
	return result.toString() ;
    } // EndMethod toString

    public String toAlphaString() {
	StringBuffer  result = new StringBuffer() ;
	result.append(MiscUtils.charCoding(min)) ;
	if (max!=min) result.append('-').append(MiscUtils.charCoding(max)) ;
	// 	    System.out.println(this + " -> " + result ) ;
	return result.toString() ;
	//  	    return toString() ;
    } // EndMethod toAlphaString

} // EndClass MinMaxOrderRange
