/*
 * Copyright 2003 LAAS-CNRS
 *
 *   LAAS-CNRS
 *   7, avenue du Colonel Roche
 *   F-31077 Toulouse Cedex 4 / France
 *
 * Author: Francois Taiani
 *
 * This file is part of OPSBrowser.
 * 
 * OPSBrowser is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * OPSBrowser is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with OPSBrowser; if not, write to the Free Software Foundation,
 * Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 * 
 */
/**
 * GenericXMLReader.java
 * 
 * Time-stamp: <2003-02-25 10:13:58 ftaiani>
 * @author Francois Taiani   <taiani@laas.fr>
 * CVS Info: $Id: GenericXMLReader.java,v 1.5 2004/09/09 15:31:01 ftaiani Exp $
 * 
 */

package laas.friends.cosmopen.graphAnalysis ;

import org.xml.sax.Attributes;

/**
 * This class reads differents formats of XML graphs
 * coded in XML, that produce {@link DirectedGraphWithDict}
 * graphs, or one of its subclasses.
 * @version $Name: release-1-3-rc1 $ $Revision: 1.5 $
 * @since Version 0.1.1
 */
public class GenericXMLReader
    extends AbstractXMLReader
{
    /**
     * The new vertices and edges are not put in a new graph,
     * but added to the <code>receiver</code> and to the <code>context</code>.
     * If both are <code>null</code>, does the same as {@link #GenericXMLReader()}
     */
    public GenericXMLReader
	(   DirectedGraphWithDict aNewReceiver    ,
	    DirectedGraphWithDict aNewContext     ) {

	receiver = aNewReceiver ; // This 2 variables are defined in AbstractXMLReader
	context  = aNewContext  ;
    }

    public GenericXMLReader() {}

    /** **********************************************
     * In {@link GenericXMLReader} this method analyse
     * the attribute of the first encountered XML tag
     * to decide which XML reader to use to analyse
     * the subsequent file. It uses the method
     * {@link AbstractXMLReader#switchToNewXMLReader(AbstractXMLReader) switchToNewXMLReader}
     * of {@link AbstractXMLReader} to dynamically change to an adapted
     * XML reader. If no known format is identified a {@link RuntimeException} is
     * thrown.
     * @see AbstractXMLReader
     * @see XMLCallGraphReader
     * @see XMLGraphReader
     * @see XMLTraceSet2CallGraphReader
     */
    public void startElement( String namespace , 
			      String localName , 
			      String qName     ,
			      Attributes someAttributes )
	throws org.xml.sax.SAXException
    {
	this.tick() ;
	
	if (localName.equals("project")) {
	    String type = someAttributes.getValue("type") ;
	    boolean isCallGraph = false ;
	    if (type!=null) isCallGraph = type.equals("callGraph") ;
	    AbstractXMLReader switchedTo ;

	    if (isCallGraph) {
		provideRecipientAndContextAccordingToGraphClass
		    ("laas.friends.cosmopen.graphAnalysis.CallGraph") ;
		switchedTo = new XMLCallGraphReader((CallGraph)receiver,(CallGraph)context);
	    }
	    else {
		provideRecipientAndContextAccordingToGraphClass
		    ("laas.friends.cosmopen.graphAnalysis.DirectedGraphWithDictImpl") ;
		switchedTo = new XMLGraphReader(receiver,context) ;
	    }

	    switchToNewXMLReader(switchedTo) ;
	    switchedTo.startElement( namespace, localName, qName, someAttributes ) ;
	}
	else if (localName.equals("traceSet")) {

	    if ((context==null) && (contextCreationListener!=null)) {
		context = CallGraph.GetCallGraphUsingChildrenRegistry();
		contextCreationListener.aGraphRecipientWasCreated(context) ;
	    } // context is null, a listener for its creation exists

	    if ((receiver==null) && (mainRecipientCreationListener!=null)) {
		receiver = CallGraph.GetCallGraphUsingChildrenRegistry();
		mainRecipientCreationListener.aGraphRecipientWasCreated(receiver) ;
	    } // receiver is null, a listener for its creation exists
	    
	    AbstractXMLReader switchedTo = new XMLTraceSet2CallGraphReader(receiver,context) ;
	    switchToNewXMLReader(switchedTo) ;
	    switchedTo.startElement( namespace, localName, qName, someAttributes ) ;
	}
	else throw new RuntimeException("Unknown tag in XML input : "+ localName );

    } // EndMethod startElement

    private void provideRecipientAndContextAccordingToGraphClass(String aGraphClassName)
	throws java.lang.RuntimeException 
    {
	try {
	    Class theGraphClass = Class.forName(aGraphClassName) ;
	
	    if ((context==null) && (contextCreationListener!=null)) {
		context = (DirectedGraphWithDict)theGraphClass.newInstance() ;
		contextCreationListener.aGraphRecipientWasCreated(context) ;
	    } // context is null, a listener for its creation exists

	    if ((receiver==null) && (mainRecipientCreationListener!=null)) {
		receiver = (DirectedGraphWithDict)theGraphClass.newInstance() ;
		mainRecipientCreationListener.aGraphRecipientWasCreated(receiver) ;
	    } // receiver is null, a listener for its creation exists
	    
	} catch(java.lang.ClassNotFoundException e){
	    throw new RuntimeException(aGraphClassName + ": ClassNotFoundException in "+
				       "provideRecipientAndContextAccordingToGraphClass") ;
	} catch(java.lang.InstantiationException e) {
	    throw new RuntimeException("InstantiationException in "+
				       "provideRecipientAndContextAccordingToGraphClass") ;
	} catch(java.lang.IllegalAccessException e) {
	    throw new RuntimeException("IllegalAccessException in "+
				       "provideRecipientAndContextAccordingToGraphClass") ;
	}

    } // EndMethod provideInstancesForGraphClassIfListenerPresent

    protected GraphRecipientCreationListener contextCreationListener       = null ;
    protected GraphRecipientCreationListener mainRecipientCreationListener = null ;

    public void setContextCreationListener(GraphRecipientCreationListener aListener) {
	contextCreationListener = aListener ;
    }

    public void setMainRecipientCreationListener(GraphRecipientCreationListener aListener){
	mainRecipientCreationListener = aListener ;
    }

} // EndClass GenericXMLReader
