/*
 * Copyright 2003 LAAS-CNRS
 *
 *   LAAS-CNRS
 *   7, avenue du Colonel Roche
 *   F-31077 Toulouse Cedex 4 / France
 *
 * Author: Francois Taiani
 *
 * This file is part of OPSBrowser.
 * 
 * OPSBrowser is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * OPSBrowser is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with OPSBrowser; if not, write to the Free Software Foundation,
 * Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 * 
 */
/**
 * NamedVertex.java
 * 
 * Time-stamp: <2002-08-16 19:40:24 ftaiani>
 * @author Francois Taiani   <taiani@laas.fr>
 * CVS Info: $Id: NamedVertex.java,v 1.8 2004/10/27 17:16:53 ftaiani Exp $
 * 
 */
package laas.friends.cosmopen.graphAnalysis  ;

import salvo.jesus.graph.Vertex ;
import java.util.Collection ;
import laas.friends.cosmopen.utils.PropertyBag ;

/**
 * This class represents a class tag extracted from an xml file
 * generated with hydrogen.
 * @author Francois Taiani   <taiani@laas.fr>
 * @version $Name: release-1-3-rc1 $ $Revision: 1.8 $
 * @since Version 0.0.1
 */
public class NamedVertex extends Vertex implements Comparable {

    private PropertyBag   _propertyContainer = new PropertyBag() ;
    private final String  _name ;
    private final int     _hashCode ;

    private static final int EVENT_ORDER_NOT_YET_COMPUTED = -2;
    private static final int EVENT_ORDER_NOT_PRESENT      = -1;
    private int eventOrderAsInt = EVENT_ORDER_NOT_YET_COMPUTED ;

    public NamedVertex () {
        _name=null ;
        _hashCode = 0 ;
    }

    public NamedVertex (String nameForNewNamedVertex) {
	_name = new String(nameForNewNamedVertex) ;
        _hashCode = _name.hashCode() ;
	_propertyContainer.setPropertyToValue("name",_name) ;
    } // EndConstructor NamedVertex(String nameForNewNamedVertex)

    public String getName() {
	return _name ;
    } // EndAccessor getName

    public int hashCode() {
        return _hashCode ;
    }

    /**
     * returns the property bag which contains
     * the properties attached to that NamedVertex
     */
    public PropertyBag getPropBag() {
	return _propertyContainer ;
    } // EndMethod getPropBag

    /**
     * returns a property value of a the {@link PropertyBag} of the
     * <code>NamedVertex</code>.
     */
    public String getValueForProperty(String aPropertyName) {
	return _propertyContainer.getValueForProperty(aPropertyName) ;
    }

    public void setPropertyToValue(String aProperty, String aValue) {
	_propertyContainer.setPropertyToValue(aProperty,aValue) ;
    } ;

    public java.lang.String toString() {
	return new String("NamedVertex : "+_name);
    } // EndMethod toString

   public final int getEventOrderAsInt() {

	if (eventOrderAsInt!=EVENT_ORDER_NOT_YET_COMPUTED) 
	    return eventOrderAsInt ;

	String  eventOrderAsString = this.getValueForProperty("eventOrder");
	if (eventOrderAsString!=null) {
	    eventOrderAsInt = Integer.parseInt(eventOrderAsString) ;
	} else
	    eventOrderAsInt=EVENT_ORDER_NOT_PRESENT ;
	return eventOrderAsInt ;
    }

    /**
     * If both compared {@link NamedVertex} possess an 'eventOrder'
     * attribute then the {@link NamedVertex} are compared according
     * to the natural number ordering on this attribute. Otherwise
     * they are compared according to the lexicagraphical order on
     * their names.
     */
    public int compareTo(Object o) {
	// [CHANGE] (Francois Taiani) Tue Oct 26 2004 20:02:11
	// parseInt is quite expensive and compare to is called *lots*
	// of time, so we need something more performant here.

// 	NamedVertex otherNamedVertex = (NamedVertex)o ;

//  	String  thisEventOrder = this            .getValueForProperty("eventOrder");
//  	String otherEventOrder = otherNamedVertex.getValueForProperty("eventOrder");

// 	if (thisEventOrder==null) {
// 	    if (otherEventOrder==null)
// 		return this.getName().compareTo(otherNamedVertex.getName()) ;
// 	    else
// 		return 1 ;
// 	} // EndIf (thisEventOrder==null)
// 	else {
// 	    if (otherEventOrder==null)
// 		return -1 ;
// 	    else
// 		return Integer.parseInt(thisEventOrder)-Integer.parseInt(otherEventOrder)  ;
// 	} // EndElse (thisEventOrder!=null)

	NamedVertex otherNamedVertex = (NamedVertex)o ;
	
	int thisEventOrder = this            .getEventOrderAsInt() ;
	int othrEventOrder = otherNamedVertex.getEventOrderAsInt();

	if (thisEventOrder == EVENT_ORDER_NOT_PRESENT ||
	    othrEventOrder == EVENT_ORDER_NOT_PRESENT )
	    
	    return this.getName().compareTo(otherNamedVertex.getName()) ;
	else
	    return thisEventOrder - othrEventOrder ;

	// The semantics is not exactly the same any more when one
	// vertex has an event order and the other one has not, but we
	// don't care that much about this case anyway.

    } // EndMethod compareTo

    /**
     * 2 {@link NamedVertex} are equals if they have the same name and
     * equals property bags.
     */
    public boolean equals(Object obj) {

        // Hashcode comparison to boost performance. According to jmp profiling,
        // this has a huge effect !

        if (this.hashCode()!=obj.hashCode()) return false ;

	if (this==obj) return true ; // trivial case
	if (!(obj instanceof NamedVertex)) return false ;


	NamedVertex autreClass = (NamedVertex)obj ;
	boolean result = this.getName().equals(autreClass.getName()) ;

	// [CHANGE] ftaiani Mon Jul  2 2001 13:27:13
	// The equals method of NamedVertex checks that the
	// properties of classes are consistent with equality.
	// If the names are equals, but the property bags
	// are not, an exception is thrown.
	    
	if ( result )
            if (!this.getPropBag().equals( autreClass.getPropBag() ) )
                throw new NamedVertexConflictException(this,autreClass) ;
	
	return result ;

    } // EndMethod equals

    /**
     * returns true if the {@link NamedVertex } has a property with
     * the name 'pseudoClass', which means that it is not a real
     * nodes, but a 'pseudo node' to represenet an n-ary relation.
     */
    public boolean isPseudo() {

	return 
	    (this
	      .getPropBag()
	      .getValueForProperty("pseudoClass")!=null) ;

    } // EndMethod isPseudo

    /**
     * Is used in {@link XMLGraphBuilder} to label the edges
     * of caller / callee types with information on the order and nesting
     * of calls. This variable counts the number of chaining between
     * incoming and outgoing calls on a given object. A chaining exist
     * if 2 calls appear in the same backtrace. If a chaining appears
     * more that one backtrace, it is counted once.
     * @see XMLGraphBuilder#addLinkFromXML( org.xml.sax.Attributes )
     * @see NamedEdge#connectToIncomingPort(String port)
     * @see NamedEdge#isAlreadyConnectedToIncomingPort(String port)
     */
    public int countOfLinkChainings = 0 ;

    /** ****************************************************
     * prints the name of the {@link NamedVertex} from a Collection. 
     * If the collection is too big, only the size of the Collection
     * is printed. The Collection is assumed to contain *only* 
     * {@link NamedVertex}, otherwise the result is not defined.
     */
    static public void PrintCollectionOfNamedVertex(Collection aCollection) {

	if (aCollection.size()<10) {
	    java.util.Iterator iter = aCollection.iterator() ;
	    while (iter.hasNext()) 
		System.out.println(((NamedVertex)iter.next()).getName()) ;
	} // EndIf (aCollection.size()<10)
	else
	    System.out.println(aCollection.size()+" nodes") ;
	
    } // EndMethod PrintCollectionOfNamedVertex

} // EndClass NamedVertex
