/*
 * Copyright 2004 Francois Taiani <f.taiani@computer.org>
 *
 * Author: Francois Taiani
 *
 * This file is part of OPSBrowser.
 * 
 * OPSBrowser is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * OPSBrowser is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with OPSBrowser; if not, write to the Free Software Foundation,
 * Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 * 
 */
/**
 * OnTheFlyCallGraphBuilder.java
 * 
 * Time-stamp: <2004-09-08 11:32:40 ftaiani>
 * @author Francois Taiani   <f.taiani@computer.org>
 * $Id: OnTheFlyCallGraphBuilder.java,v 1.6 2004/10/06 20:27:42 ftaiani Exp $
 * 
 */

package laas.friends.cosmopen.graphAnalysis  ;

import java.util.* ;

/**
 * <p>This class constructs a {@link CallGraph} on the fly. On the fly
 * construction means invocations are individually intercepted and
 * incrementally added to the graph, and there is no need to translate
 * from a sequence of stack traces to a call tree, as done in {@link
 * XMLTraceSet2CallGraphReader}. Note that the two ways of
 * construction (on the fly, and from call traces) should not be
 * mixed. The central method of this class is {@link
 * #addCallOnTheFly(String, int, String, String, String)
 * addCallOnTheFly(..)}.</p>
 *
 * <p>A tipical way to use this class would be like the following :</p>
 * <pre>
 *     OnTheFlyCallGraphBuilder builder = new OnTheFlyCallGraphBuilder() ;
 *     builder.addCallOnTheFly   ( "thread 1", 1,"A", "mA",null) ;
 *     builder.addCallOnTheFly   ( "thread 1", 2,"B", "mB",null) ;
 *     builder.registerNewThread ( "thread 2" ) ;
 *     builder.addCallOnTheFly   ( "thread 2", 1,"C", "mC",null) ;
 *     CallGraph result = builder.endGraphConstructionAndReturnResult() ;
 * </pre>
 *
 * @author Francois Taiani   <f.taiani@computer.org>
 */
public class OnTheFlyCallGraphBuilder extends CallGraphBuilder {

    public OnTheFlyCallGraphBuilder() {
        super(new CallGraph(),null); // no context graph
    } // EndConstructor OnTheFlyCallGraphBuilder

    public void addCallOnTheFly(String rawThreadID,
                                int    newFrameCount,
                                String entity, 
                                String method,
                                String thisRef 
                                ) {

        // We delegate the work to the appropriate ThreadBuilder.

//         System.out.println("addCallOnTheFly: ["+newFrameCount+"] "
//                            +entity+"::"+method+" ("+rawThreadID+")");

        ThreadBuilder thrBuilder = getBuilderForThreadID(rawThreadID) ;
        thrBuilder.addCallOnTheFly(newFrameCount, entity, method, thisRef );
 
    } // EndMethod addCallOnTheFly

    public void addThrownException(String rawThreadID, String exceptionName) {
        ThreadBuilder thrBuilder = getBuilderForThreadID(rawThreadID) ;
        thrBuilder.addThrownException(exceptionName);
        
    } // EndMethod addThrownException

    public void popFrameSForThreadID(String rawThreadID, int newFrameCount) {
        ThreadBuilder thrBuilder = getBuilderForThreadID(rawThreadID) ;
        thrBuilder.popFrameS(newFrameCount);
        
    } // EndMethod popFrameSForThreadID

    public void registerNewThread(String rawThreadID) {
        ThreadBuilder newThrBuilder = getBuilderForThreadID(rawThreadID) ;
        registerThreadCreationPoint(newThrBuilder.niceThreadID) ;
    } // EndMethod registerNewThread

    public CallGraph endGraphConstructionAndReturnResult() { 
        CallGraph result = receiver ;
        receiver = null ;
        return result ;
    } // EndMethod endGraphConstructionAndReturnResult

    private int threadCount = 0 ;
    private Map rawThreadIDtoThrdBuilder = new HashMap() ;

    public ThreadBuilder getBuilderForThreadID(String rawThreadID) {
        ThreadBuilder res = (ThreadBuilder)rawThreadIDtoThrdBuilder.get(rawThreadID);
        if (res==null) 
            res = new ThreadBuilder(rawThreadID) ;
        return res ;        
    } // EndMethod getBuilderForThreadID

    /** 
     * is responsible for the management of calls related to one given
     * thread.
     */
    private class ThreadBuilder {
        public NamedVertex currentInvok = topVertex ;
        public String      niceThreadID ;
        public int         frameCount = 0 ;

        public ThreadBuilder(String rawThreadID) {
            threadCount++ ;
            niceThreadID = Integer.toString(threadCount) ;
            rawThreadIDtoThrdBuilder.put(rawThreadID,this);
        } // EndConstructor ThreadBuilder()

        public NamedVertex popFrameS(int newFrameCount) {

            if (newFrameCount==frameCount) return currentInvok ;

            if (newFrameCount>frameCount)
                throw new RuntimeException("INTERNAL ERROR: Can't do a negative pop from "+
                                           frameCount+" to "+newFrameCount+"!") ;

            for(int i = frameCount ; i > newFrameCount ; i--) {
                currentInvok = receiver.getParent(currentInvok) ;

                // The only case when currentInvok might be null is when we pop
                // the last frame of the stack. When this happens, the
                // index i should be equal to one, and we return the
                // topVertex as a fake virtal frame.
            
                if (currentInvok==null) {
                    if (i!=1) throw new RuntimeException("INTERNAL ERROR: Null Parent") ;
                    currentInvok = topVertex ;
                }
            } // EndFor
            frameCount = newFrameCount ;
            return currentInvok ;

        } // EndMethod popFrameS

        protected void pushFrame(NamedVertex newCall) {
            currentInvok = newCall ;
            frameCount++ ;
        } // EndMethod pushFrame

        public void addCallOnTheFly(int    newFrameCount,
                                    String entity, 
                                    String method,
                                    String thisRef 
                                    ) {

            // If some function return have occurred in between we pop the
            // stack accordingly

            popFrameS(newFrameCount-1);

            EventInfo newEventInfo = new EventInfo(entity, method, thisRef, niceThreadID ) ;
            NamedVertex newInvok   = createNewEvent(newEventInfo, currentInvok) ;

            pushFrame(newInvok) ;

        } // EndMethod addCallOnTheFly

        public void addThrownException(String exceptionName) {
            addCallOnTheFly(frameCount+1, exceptionName, "throw",null) ;
        } // EndMethod addThrownException

    } // EndClass ThreadBuilder

} // EndClass OnTheFlyCallGraphBuilder
