/*
 * Copyright 2003 LAAS-CNRS
 *
 *   LAAS-CNRS
 *   7, avenue du Colonel Roche
 *   F-31077 Toulouse Cedex 4 / France
 *
 * Author: Francois Taiani
 *
 * This file is part of OPSBrowser.
 * 
 * OPSBrowser is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * OPSBrowser is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with OPSBrowser; if not, write to the Free Software Foundation,
 * Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 * 
 */
/**
 * XMLGraphReader.java
 * 
 * Time-stamp: <02/08/15 17:12:54 ftaiani>
 * @author Francois Taiani   <taiani@laas.fr>
 * CVS Info: $Id: XMLGraphReader.java,v 1.1 2004/09/09 15:34:29 ftaiani Exp $
 * 
 */
package laas.friends.cosmopen.graphAnalysis  ;

import org.xml.sax.Attributes;

/**
 * This class constructs a {@link DirectedGraphWithDictImpl } from a stream
 * containing XML data produced with Hydrogen. This data has the
 * following form: <BR>
 * <PRE>
 * &lt;class name="AB" >
 *   &lt;link kind="superclass" target="spaceOfA::includedNS::Aincluded" />
 *   &lt;link kind="usage" name="b" target="B" template="" isReference="NO" />
 *   &lt;link kind="usage" name="cIC" target="ClassInC" template="" isReference="NO" />
 * &lt;/class>
 * </PRE>
 * The main entry point of this class is {@link AbstractXMLReader#readGraphFromXML(java.io.InputStream)}.
 *
 * @author Francois Taiani   <taiani@laas.fr>
 * @version $Name: release-1-3-rc1 $ $Revision: 1.1 $
 * @since Version 0.0.1
 */

public class XMLGraphReader 
    extends AbstractXMLReader 
{
    private XMLGraphBuilder _builder ;
    private java.io.InputStream        _XMLSource ;

    /**
     * can be overwriten to change the type of builders used in the reader.
     */
    public XMLGraphBuilder createBuilder(DirectedGraphWithDict context) {
	return new XMLGraphBuilder(context) ;
    }

    /** **********************************************/
    public void initGraphConstruction() {
	// The work of incremental construction of the graph is
	// delegated to an internal XMLGraphBuilder (_builder). This
	// builder is invoked in the reification of XML parsing event in
	// XMLGraphReader.startElement(..)

	_builder = createBuilder(context) ;
	startListener("Start reading Graph from XML") ;
    }

    /** **********************************************/
    public DirectedGraphWithDict endGraphConstructionAndReturnResult() {
	DirectedGraphWithDict result = _builder.stopConstructionAndReleaseGraph() ;
	_builder = null ; // releasing builder.
	stopListener() ; 
 	return result ;
    }

    public XMLGraphReader() {
    }

    /** **********************************************
     * Does about the same as {@link #XMLGraphReader()},
     * but contrarily to it, this method adds the result of the XML
     * conversion to an already existing graph. Moreover this insertion
     * is done w.r.t to a given 'context graph': i.e. if a {@link NamedVertex }
     * is read which has the same name as a {@link NamedVertex } 
     * already present in the context graph, then no new {@link NamedVertex }
     * instance is instantiated, and the one already present in the context
     * graph is used. Note that this constructor assumed that the receiver graph
     * is contained in the context graph, from an object reference point of view:
     * all objects ({@link NamedVertex}s and {@link NamedEdge}s) present in
     * the <code>receiver</code> graph should be also present in the 
     * <code>contect</code> graph.
     */
    public XMLGraphReader
	(   DirectedGraphWithDict aNewReceiver    ,
	    DirectedGraphWithDict aNewContext     ) {

	receiver = aNewReceiver ;
	context  = aNewContext  ;
    }

    /** **********************************************
     * @see #XMLGraphReader( DirectedGraphWithDict, DirectedGraphWithDict )
     */
    public void addGraphFromXMLToGraphInContext
	(   DirectedGraphWithDict aNewReceiver    ,
	    DirectedGraphWithDict aNewContext     ,
	    java.io.InputStream   anXMLSource )
	throws 
	 javax.xml.parsers.ParserConfigurationException
	,org.xml.sax.SAXException
	,java.io.IOException
    {
	readGraphFromXML(aNewReceiver,aNewContext,anXMLSource) 
	    // will call initGraphConstruction 
	    // & endGraphConstructionAndReturnResult
	    ;
    } // EndMethod addGraphFromXMLToGraphInContext

    /** **********************************************
     * This methods reifies parsing XML events and mirrors them into
     * calls on a {@link XMLGraphBuilder}. Should no be called
     * directly.
     */
    public void startElement( String namespace , 
			      String localName , 
			      String qName     ,
			      Attributes someAttributes )
    {
	this.tick() ;
	
	if (localName.equals("class"))

	    _builder.startAddingNamedVertexFromXML(someAttributes) ;

	else if (localName.equals("link"))

	    _builder.addLinkFromXML( someAttributes ) ;

	else if (localName.equals("project"))

	    _builder.startGraphConstruction(receiver,someAttributes) ;

	else throw new RuntimeException("Unknown tag in XML input : "+ localName ) ;

    } // EndMethod startElement

    /** **********************************************/
    public void endElement( String namespace, 
			    String localName, 
			    String qName ) {
	
	if (localName.equals("class"))
	    _builder.endAddingNamedVertexFromXML() ;
	else if (localName.equals("link"))
	    ;
	else if (localName.equals("project"))
	    ;
	else throw new RuntimeException("Unknown tag in XML input : "+ localName ) ;

    } // EndMethod endElement

} // EndClass XMLGraphReader
