/*
 * Copyright 2003 LAAS-CNRS
 *
 *   LAAS-CNRS
 *   7, avenue du Colonel Roche
 *   F-31077 Toulouse Cedex 4 / France
 *
 * Author: Francois Taiani
 *
 * This file is part of OPSBrowser.
 * 
 * OPSBrowser is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * OPSBrowser is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with OPSBrowser; if not, write to the Free Software Foundation,
 * Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 * 
 */
/**
 * XMLTraceSet2CallGraphReader.java
 * 
 * Time-stamp: <2003-03-10 19:11:04 ftaiani>
 * @author Francois Taiani   <taiani@laas.fr>
 * CVS Info: $Id: XMLTraceSet2CallGraphReader.java,v 1.10 2004/10/06 20:27:42 ftaiani Exp $
 * 
 */

package laas.friends.cosmopen.graphAnalysis  ;

import org.xml.sax.Attributes;
import java.io.* ;
import laas.friends.cosmopen.utils.PropertyBag ;

/**
 * This class builds a {@link CallGraph} from a
 * set of traces in XML format.
 *
 * @author Francois Taiani   <taiani@laas.fr>
 * @version $Name: release-1-3-rc1 $ $Revision: 1.10 $
 * @since Version 0.1.1
 */
public class XMLTraceSet2CallGraphReader
    extends AbstractXMLReader
{

    public XMLTraceSet2CallGraphReader() {}

    public XMLTraceSet2CallGraphReader
	(   DirectedGraphWithDict aNewReceiver    ,
	    DirectedGraphWithDict aNewContext     ) {

	receiver = aNewReceiver ;
	context  = aNewContext  ;
    }

    private CallGraphBuilder builder ;

    public void initGraphConstruction() {

	if ( receiver==null ) receiver = CallGraph.GetCallGraphUsingChildrenRegistry() ;

        builder        = new CallGraphBuilder((CallGraph)receiver,(CallGraph)context) ;
        previousParent = builder.topVertex ;
	startListener("Start reading trace graph from XML") ;
    }

    public DirectedGraphWithDict endGraphConstructionAndReturnResult() {
	stopListener() ; 
	((CallGraph)receiver).switchOffUseOfChildrenRegistry() ;
 	return receiver ;
    }

    /**
     * This methods reifies parsing XML events and mirrors them into
     * calls on a {@link XMLGraphBuilder}. Should no be
     * called directly.
     */
    public void startElement( String namespace , 
			      String localName , 
			      String qName     ,
			      Attributes someAttributes )
    {
	this.tick() ;
// 	System.out.println(localName+" : "
// 			   +someAttributes.getValue("entity")+"::"
// 			   +someAttributes.getValue("method")) ;
	if (localName.equals("trace"))
	    handleBeginningOfTrace(someAttributes) ;
	else if (localName.equals("call"))
	    addCallEntry(someAttributes, currentStackThread) ;
	else if (localName.equals("new")) {
            if (weAreWithinATraceTag)
                throw new RuntimeException
                    ("Found a thread creation <new> point within a <trace></trace> tag.") ;
	    builder.registerThreadCreationPoint(someAttributes.getValue("thread")) ;
        }
	else if (localName.equals("end")) {
            if (weAreWithinATraceTag)
                throw new RuntimeException
                    ("Found a thread creation <end> point within a <trace></trace> tag.") ;
	    builder.registerThreadTerminationPoint(someAttributes.getValue("thread")) ;
        }
	else if (localName.equals("traceSet"))
	    handleBeginningOfTraceSet(someAttributes) ;
	else throw new RuntimeException("Unknown tag in XML input : "+ localName ) ;

    } // EndMethod startElement

    public void endElement( String namespace, 
			    String localName, 
			    String qName ) {
	
	if (localName.equals("trace"))
	    handleEndOfTrace() ;
	else if (localName.equals("call"))     ;
	else if (localName.equals("new"))      ;
	else if (localName.equals("end"))      ;
	else if (localName.equals("traceSet")) ;
	else throw new RuntimeException("Unknown tag in XML input : "+ localName ) ;

    } // EndMethod endElement

    /** used to handle thread creation point */

    private String          currentStackThread ;
    private java.util.Stack stackOfCalls ;
    private boolean         weAreWithinATraceTag = false ;

    /**
     * In some traces, the thread creation as registered by gdb
     * is delayed. Because by default the graph builder consider
     * the top of the last stack trace as being the point of
     * thread creation, wrong thread creation links are created.

     * <pre>
     *   <trace thread="2" > 
     *     <call entity="" method="clone" this="" />
     *     <call entity="" method="pthread_handle_create" this="" />
     *     <call entity="" method="__pthread_manager" this="" />
     *     <call entity="" method="pthread_start_thread" this="" />
     *   </trace>
     *   <trace thread="4" > 
     *     <call entity="" method="gettimeofday" this="" />
     *     <call entity="" method="__pthread_timedsuspend_new" this="" />
     *     <call entity="" method="pthread_cond_timedwait_relative" this="" />
     *     <call entity="" method="pthread_cond_timedwait" this="" />
     *     <call entity="" method="pthread_start_thread_event" this="" />
     *   </trace>
     *   <new thread="5" />
     * </pre>

     * For instance in the previous trace extract, the graph builder
     * considers that thread 5 was created within gettimeofday, which 
     * is wrong.<p>
     * In order to avoid that, it's possible to add an "threadCreation" attribute
     * to the <traceSet> tag that indicate likely thread creation points.
     * (here "clone"). The last encountered likeley thread creation point
     * is considered to find the point where the thread was created.
     */
    private void handleBeginningOfTraceSet(Attributes someAttributes) {

	builder.setUsualThreadCreationPoint( someAttributes.getValue("threadCreation") ) ;
	// returns null if not defined
	
    } // EndMethod handleBeginningOfTrace

    private void handleBeginningOfTrace(Attributes someAttributes) {
	stackOfCalls       = new java.util.Stack() ;
	currentStackThread = someAttributes.getValue("thread") ;
	weAreWithinATraceTag = true ;
    }

    private void addCallEntry(Attributes someAttributes, String currentStackThread) {
	CallGraphBuilder.EventInfo newCall 
            = builder.new EventInfo(someAttributes, currentStackThread) ;
	stackOfCalls.push(newCall) ;
    } // EndMethod addCallEntry

    private NamedVertex previousParent = null ;

    /** **************************************************** */
    private void handleEndOfTrace() {

	weAreWithinATraceTag = false ;

	// (Francois Taiani) Fri Nov 29 2002 10:13:55
	// If we don't do that, the program records the topVertex
	// as being the lastCallOfLastTrace, which totally confuses
	// the code that handles "thread creation" links.
	if (stackOfCalls.empty()) return ;

        // alreadyInTree = false ;
	previousParent = builder.topVertex ;
// 	eventID   = null ;

	while (!stackOfCalls.empty() ) {

	    ChildrenRegistry currentChildrenRegistry 
		= builder.getChildrenRegistryForVertex(previousParent) ;
	    CallGraphBuilder.EventInfo currentCall = 
                (CallGraphBuilder.EventInfo)stackOfCalls.pop() ;
// 	    String   callPoint   = ChildrenRegistry.getNameOfCallPoint
// 		( currentCall.callAttributes.getValueForProperty("entity"),
// 		  currentCall.callAttributes.getValueForProperty("method")
// 		  ) ;

// 	    System.out.println("Handling callPoint:" + callPoint) ;

	    // [CHANGE] (Francois Taiani) Thu Mar  6 2003 09:27:44
	    // Correcting bug regarding wrong interpretation of trace sets that
	    // contain cycling behavior: recurring invocation of the same method
	    // were not considered as distinct events with the previous implementation.
	    //
	    //	    if (!currentChildrenRegistry
	    //           .doesContainChildWithCallPoint(callPoint)) {
	    //
	    // [CHANGE] (Francois Taiani) Fri Mar  7 2003 20:04:13
	    // stackOfCalls.empty() to enable of distinction of successive
	    // breakpoints with exactly the same stack trace: in this
	    // particular case we know the terminal stack frame correspond
	    // to distinct invocations.

	    if (stackOfCalls.empty() || 
		!currentChildrenRegistry
		.youngestChildForThreadHasSameCallPoint(currentStackThread,
                                                        currentCall.getCallPointString())) 
                {
                    // the call point is not the same as the one of
                    // the last registered invocation for this
                    // thread. So this is a new invocation.

                    previousParent = builder.createNewEvent(currentCall, previousParent) ;
                }
	    else { 
		// The call point is not new. We don't need to create
		// any new event

                boolean alreadyInTree = (previousParent!=builder.topVertex) ;

		previousParent = 
                    currentChildrenRegistry.getYoungestChildForThread(currentStackThread);

		if (!alreadyInTree) {

		    String  foundThread   = previousParent.getValueForProperty("threadID") ;
		    boolean inconsistency = false ;
		    if (foundThread==null)
			inconsistency = (currentStackThread!=null) ;
		    else
			inconsistency = !foundThread.equals(currentStackThread) ;
		    if (inconsistency)
			throw 
			    (new RuntimeException
			     ("INTERNAL ERROR in XMLTraceSet2CallraphReader.handleEndOfTrace : "+
			      "Thread IDs don't match. " + foundThread +
			      " <> " + currentStackThread ) ) ;
		} // EndIf we're at the top of the tree

	    } // End Else The call point is not new

// 	    alreadyInTree=true ;

	} // EndWhile (!stackOfCalls.empty() )

// 	lastCallOfLastTrace = previousParent ;

    } // EndMethod handleEndOfTrace

    /** ****************************************************
     * Transparently creates a {@link XMLTraceSet2CallGraphReader}
     * and uses it to read a CallGraph from a string representing
     * a set of traces.
     */
    public static CallGraph ReadCallGraphFromString(String aGraphInString) 
	throws javax.xml.parsers.ParserConfigurationException,
	       org.xml.sax.SAXException,
	       java.io.IOException
    {
	XMLTraceSet2CallGraphReader reader = new XMLTraceSet2CallGraphReader() ;

 	CallGraph result ;
 	result = 
	    (CallGraph)reader
	    .readGraphFromXML(new ByteArrayInputStream(aGraphInString.getBytes())) ;

	return result ;
    } // Method ReadCallGraphFromString

} // EndClass XMLTraceSet2CallGraphReader
