/*
 * Copyright 2003 LAAS-CNRS
 *
 *   LAAS-CNRS
 *   7, avenue du Colonel Roche
 *   F-31077 Toulouse Cedex 4 / France
 *
 * Author: Francois Taiani
 *
 * This file is part of OPSBrowser.
 * 
 * OPSBrowser is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * OPSBrowser is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with OPSBrowser; if not, write to the Free Software Foundation,
 * Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 * 
 */
/**
 * AbstractDocumentedItem.java
 * 
 * Time-stamp: <2003-02-16 17:54:07 ftaiani>
 * @author Francois Taiani   <taiani@laas.fr>
 * $Id: AbstractDocumentedItem.java,v 1.3 2005/04/13 16:23:44 ftaiani Exp $
 * 
 */

package laas.friends.cosmopen.interpreter;

import java.util.* ;

/**
 *  <code>AbstractDocumentedItem</code> is the common parent
 * class of {@link Command} and {@link GroupOfDocumentedItemS}. Those
 * three classes implements a "composite pattern" as defined
 * in the GoF book on "Design Pattern".
 * Basically any instance of <code>AbstractDocumentedItem</code>
 * can provide a description of itself, both in short and long
 * format.
 * @author Francois Taiani   <taiani@laas.fr>
 */
public abstract class AbstractDocumentedItem implements Comparable {

    private String _name ;
    private String _shortDescription = null ;
    private String _longDescription  = null ;
    private CommandInterpreter _myCommandInterpreter = null ;

    public AbstractDocumentedItem(String aDocumentedItemName) {
	_name = aDocumentedItemName ;
   } // EndConstructor AbstractDocumentedItem

    public void setContainingInterpreter(CommandInterpreter someInterpreter) {
	_myCommandInterpreter = someInterpreter ;
    } // EndMethod setContainingInterpreter
    
    public CommandInterpreter containingInterpreter() {
	return _myCommandInterpreter ;
    } // EndMethod containingInterpreter

    public String getName() {
	return _name;
    } // EndMethod getName

    public void setShortDescription(String aShortDescription) {
	_shortDescription = aShortDescription ;
    } // EndMethod setShortDescription

    public void setLongDescription(String aLongDescription) {
	_longDescription = aLongDescription ;
    } // EndMethod setShortDescription

    public String getShortDescription() {
	if (_shortDescription!=null) return _shortDescription ;
	else                         return "not documented" ;
    } // EndMethod getShortDescription

    public boolean hasLongDescription() { return (_longDescription!=null) ; }

    public String getLongDescription() {
	if (_longDescription!=null)  return _longDescription ;
	else                         return _name + " is not documented." ;
    } // EndMethod getLongDescription

    /** ****************************************************
     * By default, returns the same as getLongDescription, but
     * this may be changed. This method is called when help
     * is required on the corresponding item.
     * @see GroupOfDocumentedItemS#getHelpOutput()
     * @see CommandInterpreter#printHelpForItem(String)
     */
    public String getHelpOutput() {
	return getLongDescription() ;
    } // EndMethod getHelpOutput

    /** ****************************************************
     * Returns the bullet that should be used in a list.
     * The idea is to used different bullets for
     * {@link Command}s and {@link GroupOfDocumentedItemS}.
     */
    public char getBullet() {
	return '-' ;
    }

    /** ****************************************************
     * By default a {@link AbstractDocumentedItem} does not
     * support aliasing, so <code>getAliasES</code> simply
     * returns a set containing its name. Doing so make
     * handling {@link Command}s and {@link GroupOfDocumentedItemS}
     * uniformely easier.
     */
    public SortedSet getAliasES() {
	SortedSet result = new TreeSet() ;
	result.add(getName()) ;
	return result ;
    }

    public int compareTo(Object o) {
	AbstractDocumentedItem other = (AbstractDocumentedItem)o ;
	return (this.getName().compareTo(other.getName())) ;
    } // EndMethod compareTo

    public String toString() {
	return super.toString() + ":\"" + getName() + "\"" ;
    } // EndMethod compareTo

} // EndClass AbstractDocumentedItem
