/*
 * Copyright 2003 LAAS-CNRS
 *
 *   LAAS-CNRS
 *   7, avenue du Colonel Roche
 *   F-31077 Toulouse Cedex 4 / France
 *
 * Author: Francois Taiani
 *
 * This file is part of OPSBrowser.
 * 
 * OPSBrowser is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * OPSBrowser is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with OPSBrowser; if not, write to the Free Software Foundation,
 * Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 * 
 */
/**
 * TestCommandInterpreter.java
 * 
 * Time-stamp: <2003-03-11 11:14:18 ftaiani>
 * @author Francois Taiani   <taiani@laas.fr>
 * CVS Info: $Id: TestCommandInterpreter.java,v 1.4 2004/09/09 15:31:02 ftaiani Exp $
 * 
 */
package laas.friends.cosmopen.interpreter  ;

import junit.framework.TestCase ;
import java.util.* ;
import java.io.* ;

/**
 * Regresion Tests for {@link CommandInterpreter }.
 * @author Francois Taiani   <taiani@laas.fr>
 * @version $Name: release-1-3-rc1 $ $Revision: 1.4 $
 * @since Version 0.0.2
 */
public class TestCommandInterpreter
    extends AbstractCommandInterpreterTest {

    public TestCommandInterpreter(String name) {
	super(name) ;
    } // EndConstructor TestCommandInterpreter

    public void testCommandInterpreterBasics() 	
	throws Exception
    {
	CommandInterpreter testee = new CommandInterpreter("Testee");

	// Commands 'help' and 'quit' and 'run' should exist
	// by default. Empty string should be ignored.

	try {
	    testee.interpretCommand("help") ;
	    testee.interpretCommand("quit") ;
	    testee.interpretCommand("run" ) ;
	    testee.interpretCommand("" ) ;
	} catch (Exception e) {
	    e.printStackTrace() ;
	    fail() ;
	} 

	// Call on a command which is not known should raise an
	// UnknownItemException

	try {
	    testee.interpretCommand("DoesNotExist") ;
	    fail() ;
	} catch (UnknownItemException e) {
	} 

	TestingCommand aNewCommand = 
	    new TestingCommand("NewCommand", this ) ;
	testee.addCommandToKnownOnes(aNewCommand) ;
	String [] expectedArgs ;

	try {

	    // TESTS CALL WITH NO ARGUMENTS
	    expectedArgs = new String[] {"NewCommand"} ; // no args
	    aNewCommand.initForInterpretation(expectedArgs) ;
	    testee.interpretCommand("NewCommand") ;
	    aNewCommand.checkHasBeenCalled() ;
	    
	    // TESTS CALL WITH 2 ARGUMENTS
	    expectedArgs = new String[] {"NewCommand","Arg1"} ;
	    aNewCommand.initForInterpretation(expectedArgs) ;
	    testee.interpretCommand("NewCommand Arg1") ;
	    aNewCommand.checkHasBeenCalled() ;
	    testee.interpretCommand("NewCommand    Arg1") ;
	    aNewCommand.checkHasBeenCalled() ;
	    testee.interpretCommand("NewCommand    Arg1   ") ;
	    aNewCommand.checkHasBeenCalled() ;

	    // TESTS CALL WITH 4 ARGUMENTS
	    expectedArgs = 
		new String[] {"NewCommand","Arg1","Arg2","XXXXXXXXX","42"} ;
	    aNewCommand.initForInterpretation(expectedArgs) ;
	    testee.interpretCommand("NewCommand Arg1 Arg2 XXXXXXXXX 42") ;
	    aNewCommand.checkHasBeenCalled() ;
	    testee.interpretCommand("NewCommand    Arg1  Arg2 XXXXXXXXX    42") ;
	    aNewCommand.checkHasBeenCalled() ;

	} catch (UnknownItemException e) {
	    e.printStackTrace() ;
	    fail() ;
	} 

    } // EndMethod testCommandInterpreterBasics

    public void testInterpretCommandStream() throws Exception {
	
	CommandInterpreter testee = new CommandInterpreter("Testee");
	String [] expectedArgs ;
	TestingCommand command1 = new TestingCommand("Command1", this ) ;
	TestingCommand command2 = new TestingCommand("Command2", this ) ;
	testee.addCommandToKnownOnes(command1) ;
	testee.addCommandToKnownOnes(command2) ;

	expectedArgs = new String[] {"Command1","arg1"} ;
	command1.initForInterpretation(expectedArgs) ;
	expectedArgs = new String[] {"Command2"} ; // no args
	command2.initForInterpretation(expectedArgs) ;

	String stimulus = "Command1 arg1 \n Command2" ;
	testee.interpretCommandStream( new ByteArrayInputStream(stimulus.getBytes())) ;

	command1.checkHasBeenCalled() ;
	command2.checkHasBeenCalled() ;

    } // EndMethod testInterpretCommandStream

    public void testArgNumberCheck() {

	String[] args0 = new String[] {"foo"} ;
	String[] args1 = new String[] {"foo","arg"} ;
	String[] args2 = new String[] {"foo","arg","arg"} ;

	Command testee = new CommandWithCheck("foo") ;
	assertTrue(   testee.thereIsRightNumberOfArgs(args0)) ;
	assertTrue( ! testee.thereIsRightNumberOfArgs(args1)) ;

	testee = new CommandWithCheck("foo",0,0) ;
	assertTrue(   testee.thereIsRightNumberOfArgs(args0)) ;
	assertTrue( ! testee.thereIsRightNumberOfArgs(args1)) ;

	testee = new CommandWithCheck("foo",0,1) ;
	assertTrue(   testee.thereIsRightNumberOfArgs(args0)) ;
	assertTrue(   testee.thereIsRightNumberOfArgs(args1)) ;
	assertTrue( ! testee.thereIsRightNumberOfArgs(args2)) ;

	testee = new CommandWithCheck("foo",1,1) ;
	assertTrue( ! testee.thereIsRightNumberOfArgs(args0)) ;
	assertTrue(   testee.thereIsRightNumberOfArgs(args1)) ;
	assertTrue( ! testee.thereIsRightNumberOfArgs(args2)) ;

	testee = new CommandWithCheck("foo",1,2) ;
	assertTrue( ! testee.thereIsRightNumberOfArgs(args0)) ;
	assertTrue(   testee.thereIsRightNumberOfArgs(args1)) ;
	assertTrue(   testee.thereIsRightNumberOfArgs(args2)) ;

     } // EndMethod testArgNumberCheck

    static private boolean CompareStringArray( String[] array1,
					       String[] array2 )
    {
	if (array1.length!=array2.length) return false ;
	for( int i=0; i < array1.length; i++) {
// 	    System.out.print("i = ") ;
// 	    System.out.print(i) ;
// 	    System.out.println(" a1 "+array1+" a2 "+array2);
	    if (!array1[i].equals(array2[i])) return false;
	} // EndFor Loop through arrays
	return true ;
    } // EndStaticMethod CompareStringArray

    /**
     * Is used in {@link #testArgNumberCheck() }.
     */
    private class CommandWithCheck extends Command {
	public CommandWithCheck(String aCommandName) {
	    super(aCommandName) ;
	} // EndConstructor CommandWithCheck

	public CommandWithCheck(String aCommandName, 
				int    minNumbOfArgs,
				int    maxNumbOfArgs) {
	    super(aCommandName,minNumbOfArgs,maxNumbOfArgs) ;
	} // EndConstructor CommandWithCheck

	// Dummy Implementation
	public boolean handleCall(String[] args) { return true ;} 

    } // EndInnerClass CommandWithCheck

    /**
     * is used in {@link #testCommandInterpreterBasics() }. This
     * class contains a flag {@link #_hasBeenCalled } which
     * is flipped once {@link #handleCall(String[] args) } has
     * actually been called. {@link #checkHasBeenCalled() } fails
     * the test in case after an interpretation through
     * a {@link CommandInterpreter } this flag has not been flipped.
     */
    private class TestingCommand extends Command {
	private junit.framework.TestCase _callingTestCase ;
	private String[]                 _referenceArguments ;
	private	boolean                  _hasBeenCalled ;

	public TestingCommand( String   aCommandName ) {
	    super(aCommandName) ;
	}

	public TestingCommand( String   aCommandName, 
			       TestCase aCallingTestCase) 
	{
	    super(aCommandName) ;
	    _callingTestCase = aCallingTestCase ;
	} // EndConstructor

	public boolean handleCall(String[] args) {
	    _hasBeenCalled = true ;
	    _callingTestCase.assertTrue
		("Passed and received arguments are different for "+getName(),
		 CompareStringArray(_referenceArguments,args)
		 ) ;
	    return true ;
	} // EndMethod handleCall

	public void initForInterpretation(String[] expectedArgs) {
	    _referenceArguments = expectedArgs ;
	    _hasBeenCalled      = false ;
	} // EndMethod initForInterpretation

	/**
	 * fails if flag {@link #_hasBeenCalled } is false.
	 * If true, sets it thean back to false.
	 */
	public void checkHasBeenCalled() {
	    _callingTestCase.assertEquals
		("Command "+getName()+" has not been called." ,
		 true,_hasBeenCalled) ;
	    _hasBeenCalled = false ;
	} // EndMethod initForInterpretation

    } // EndPrivateClass testingCommand


    /** **************************************************** */
    public void testCompleteNameMethods() {
	CommandInterpreter testee = new CommandInterpreter("Testee");

	Command uneCommande = new TestingCommand("uneCommande" ) ;
	Command junk1       = new TestingCommand("junk1"       ) ;
	Command junk2       = new TestingCommand("junk2"       ) ;
	Command alphajunk   = new TestingCommand("alphajunk"   ) ;

	testee.addCommandToKnownOnes(uneCommande) ;
	testee.addCommandToKnownOnes(junk1      ) ;
	testee.addCommandToKnownOnes(junk2      ) ;
	testee.addCommandToKnownOnes(alphajunk  ) ;

	Collection result ;
	result = testee.completeCommandName("doesNotExist") ;
	assertEquals(new TreeSet(),result);
	result = testee.completeItemName("doesNotExist") ;
	assertEquals(new TreeSet(),result);

	Collection expected ;

	expected = new TreeSet() ;
	expected.add(uneCommande) ;
	result = testee.completeCommandName("uneCommande") ;

//     	System.out.println("\nEXPECTED\n\n" + expected ) ;
//     	System.out.println("\nRESULT  \n\n" + result ) ;

	assertEquals(expected,result);

	result = testee.completeItemName("uneCommande") ;
	assertEquals(expected,result);

	testee.addCommandToKnownOnes(new TestingCommand("uneCommandeQuiEnFaitPlus")) ;

	result = testee.completeCommandName("uneCommande") ;

//     	System.out.println("\nEXPECTED\n\n" + expected ) ;
//     	System.out.println("\nRESULT  \n\n" + result ) ;

	assertEquals(expected,result);

	result = testee.completeItemName("uneCommande") ;
	assertEquals(expected,result);

	expected = new TreeSet() ;
	expected.add(junk1) ;
	expected.add(junk2) ;
	result = testee.completeCommandName("junk") ;

//     	System.out.println("\nEXPECTED\n\n" + expected ) ;
//     	System.out.println("\nRESULT  \n\n" + result ) ;

	assertEquals(expected,result);

	result = testee.completeItemName("junk") ;
	assertEquals(expected,result);
	
	// We now test that groups are taken into account
	// Groups should appear when doing a completeItemName, but
	// not in a completeCommandName

	GroupOfDocumentedItemS junk3Group   = new GroupOfDocumentedItemS("junk3") ;
	testee.addGroupToKnownOnes(junk3Group) ;
	Collection oldExpected = new TreeSet(expected) ;
	expected.add(junk3Group) ;

	assertEquals(oldExpected,testee.completeCommandName("junk"));
	assertEquals(   expected,testee.completeItemName   ("junk"));

	GroupOfDocumentedItemS junkGroup   = new GroupOfDocumentedItemS("junk") ;
	testee.addGroupToKnownOnes(junkGroup) ;
	expected = new TreeSet() ;
	expected.add(junkGroup) ;

	assertEquals(oldExpected,testee.completeCommandName("junk"));
	assertEquals(   expected,testee.completeItemName   ("junk"));

	// COMMAND WITH ALIAS: SHOULD BEHAVE THE SAME

	Command wunk1       = new TestingCommand("wunk1") ;
	Command wunk2       = new TestingCommand("wunk2") ;
 	wunk1.addAlias("wx");
	wunk2.addAlias("wu");

	testee.addCommandToKnownOnes(wunk1) ;
	testee.addCommandToKnownOnes(wunk2) ;

	expected = new TreeSet() ;
	expected.add(wunk1) ;
	expected.add(wunk2) ;

	result = testee.completeCommandName("wunk") ;
	assertEquals("testee.completeCommandName(\"wunk\")",expected,result);

	result = testee.completeItemName("wunk") ;
	assertEquals("testee.completeItemName(\"wunk\")",expected,result);

	// WHEN COMPLETED STRING IS AN ALIAS: ONLY CORRESPONDING
	// COMMAND SHOULD BE RETURNED

	expected = new TreeSet() ;
	expected.add(wunk2) ;

	result = testee.completeCommandName("wu") ;
	assertEquals("testee.completeCommandName(\"wu\")",expected,result);

	result = testee.completeItemName("wu") ;
	assertEquals("testee.completeItemName(\"wu\")",expected,result);

    } // EndMethod testCompleteNameMethods


    /** **************************************************** */
    class ObservableInterpreter extends CommandInterpreter {
	public String  lastCommandSentToShell = null ;
	public boolean lastTerminationHint    = false ;

	public ObservableInterpreter(String interpreterName) {
	    super(interpreterName) ;
	} ; // EndConstructor

	protected void executeCommandAndPossiblyWait(String cmdString, boolean waitForTermination ) 
	    throws IOException, InterruptedException 
	{
	    lastCommandSentToShell = cmdString ;
	    lastTerminationHint    = waitForTermination ;
	} // EndMethdod executeCommandAndPossiblyWait

    } ; // EndSubclass ObservableInterpreter

    /** **************************************************** */
    public void testShellDelegation() throws Exception {

	// We test here that any command line that begins with !
	// is forwarded to the underlying shell. To do that we need to 
	// subclas the original interpreter class.

	ObservableInterpreter testee = new ObservableInterpreter("testee");
	String commandString = "this is a command shell with args" ;
	testee.interpretCommand("!"+commandString);
	assertEquals(commandString,testee.lastCommandSentToShell) ;
	assertEquals(commandString,true,testee.lastTerminationHint) ;
	
	testee.interpretCommand("!"+commandString+"&");
	assertEquals(commandString,testee.lastCommandSentToShell) ;
	assertEquals(commandString,false,testee.lastTerminationHint) ;

	testee.interpretCommand("!"+commandString+" & ");
	assertEquals(commandString,testee.lastCommandSentToShell) ;
	assertEquals(commandString,false,testee.lastTerminationHint) ;

	String[] result ;
	String[] expected = { "!" } ;

	result = testee.breakStringIntoArgs("!") ;

	assertEquals("Testing breaking '!'"  ,expected.length,result.length) ;
	assertEquals("Testing breaking '!'"  ,expected[0]    ,result[0]    ) ;
	result = testee.breakStringIntoArgs("! ") ;
	assertEquals("Testing breaking '! '" ,expected.length,result.length) ;
	assertEquals("Testing breaking '! '" ,expected[0]    ,result[0]    ) ;
	result = testee.breakStringIntoArgs(" ! ") ;
	assertEquals("Testing breaking ' ! '",expected.length,result.length) ;
	assertEquals("Testing breaking ' ! '",expected[0]    ,result[0]    ) ;

    } // EndMethod testShellDelegation


    /** **************************************************** */
    public void testCommandGrouping() {
	CommandInterpreter testee = new CommandInterpreter("Testee");
	GroupOfDocumentedItemS topGroup = testee.getTopDocumentationGroup() ;
	// The top documentation group should contain a group named
	// general that contains the default commands.
	SortedSet contentOfTopGroup = topGroup.getGroupContent() ;
	Iterator cursor = contentOfTopGroup.iterator() ;
	GroupOfDocumentedItemS general = null ;
	while((cursor.hasNext()) && (general==null)) {
	    AbstractDocumentedItem currentItem = (AbstractDocumentedItem)cursor.next() ;
	    if (currentItem.getName().equals("general"))
		general = (GroupOfDocumentedItemS)currentItem ;
	} // EndWhile
	assertTrue("Grouping A",general!=null) ;
	//	assertTrue("Grouping B",topGroup.getDocumentedItemWithName("General")==general) ;
	SortedSet contentOfGeneral = general.getGroupContent() ;

	assertTrue("Grouping C",contentOfGeneral.contains(testee.getItemForName("help"))) ;
	assertTrue("Grouping D",contentOfGeneral.contains(testee.getItemForName("quit"))) ;
	assertTrue("Grouping E",contentOfGeneral.contains(testee.getItemForName("run" ))) ;
	assertTrue("Grouping F",contentOfGeneral.contains(testee.getItemForName("!"   ))) ;

	TestingCommand aNewCommand = 
	    new TestingCommand("NewCommand") ;

	// When added with no group mentionned, the command is added to the 
	// general group.

	testee.addCommandToKnownOnes(aNewCommand) ;

	assertTrue("Grouping G",general.getGroupContent().contains(aNewCommand)) ;
	assertTrue("Grouping H",aNewCommand==testee.getItemForName("NewCommand")) ;
	
	TestingCommand aNewCommand2 = 
	    new TestingCommand("NewCommand2") ;
	GroupOfDocumentedItemS aNewGroup =
	    new GroupOfDocumentedItemS("a new group" ) ;
	
	// Adding a command to an unknown group raises an exception
	try {
	    testee.addItemToKnownOnesIntoGroup(aNewCommand2,aNewGroup) ;
	    fail() ;
	} catch (RuntimeException e) {}

	// After adding a group and a command to that group, the group
	// and the command are known to the interpreter.

	testee.addGroupToKnownOnes(aNewGroup) ;
	testee.addItemToKnownOnesIntoGroup(aNewCommand2,aNewGroup) ;

	assertTrue("Grouping I", aNewGroup    == testee.getItemForName("a new group") ) ;
	assertTrue("Grouping J", aNewCommand2 == testee.getItemForName("NewCommand2") ) ;

	Collection expected = new TreeSet() ;
	expected.add(aNewCommand ) ;
	expected.add(aNewCommand2) ;

	Collection result = testee.completeCommandName("NewCom") ;
	assertEquals("Grouping J1",expected,result) ;

	// And the command can be found in the elements of the
	// group.

	assertTrue("Grouping K", !general.getGroupContent().contains(aNewCommand2)) ;
	assertTrue("Grouping L",aNewGroup.getGroupContent().contains(aNewCommand2)) ;
	
	TestingCommand aCloneOfNewCommand = 
	    new TestingCommand("NewCommand") ;

	// Adding a command with the same name as an existing one
	// to a different group raises an exception

	try {
	    testee.addItemToKnownOnesIntoGroup(aCloneOfNewCommand,aNewGroup) ;
	    fail() ;
	} catch (RuntimeException e) {}

	GroupOfDocumentedItemS aCloneOfNewGroup =
	    new GroupOfDocumentedItemS("a new group" ) ;
	
	// Idem for a group
	try {
	    testee.addGroupToKnownOnes(aCloneOfNewGroup) ;
	    fail() ;
	} catch (RuntimeException e) {}

	// A group can be added recursively be added to a group,
	// and a command can be added to that subgroup

	GroupOfDocumentedItemS aSubgroupOfNewGroup =
	    new GroupOfDocumentedItemS("subgroup" ) ;

	testee.addItemToKnownOnesIntoGroup(aSubgroupOfNewGroup,aNewGroup) ;
	assertTrue("Grouping M", aSubgroupOfNewGroup == testee.getItemForName("subgroup") ) ;
	assertTrue("Grouping N",aNewGroup.getGroupContent().contains(aSubgroupOfNewGroup)) ;

	TestingCommand yetAnotherCommand = 
	    new TestingCommand("yetAnotherCommand") ;

	testee.addItemToKnownOnesIntoGroup(yetAnotherCommand,aSubgroupOfNewGroup) ;
	assertTrue("Grouping O", yetAnotherCommand == testee.getItemForName("yetAnotherCommand") ) ;
	assertTrue("Grouping P", aSubgroupOfNewGroup.getGroupContent().contains(yetAnotherCommand)) ;

	// It should be possible to add the same command to different group.

	testee.addItemToKnownOnesIntoGroup(yetAnotherCommand,aNewGroup) ;
	assertTrue("Grouping Q", yetAnotherCommand == testee.getItemForName("yetAnotherCommand") ) ;
	assertTrue("Grouping R", aNewGroup.getGroupContent().contains(yetAnotherCommand)) ;

    } // EndMethod testCommandGrouping

} // EndClass TestCommandInterpreter
