package salvo.jesus.graph;

import java.util.*;

/**
 * The DirectedAcyclicGraph class represents a directed acyclic graph (DAG)
 * where there is no cyclic paths for any of its vertices.
 *
 * A cylic path is a path from a vertex back to itself by following the
 * direction of the edges.
 *
 * @author		Jesus M. Salvo Jr.
 */

public interface DirectedAcyclicGraph extends DirectedGraph {

  /**
    * Returns a Vector of vertices that is not depended on by other vertices.
    * That is, a Vector of vertices where there are no Edges pointing to it.
    *
    * @return	Vector of vertices
    */
  public Collection getRoot( );

  /**
   *  Perform a topological sort of the entire directed acyclic graph.
   *  Note that the sequence of vertices in the return Vector will not distinguish
   *  between connected components of the graph.
   *
   *  @return Vector containing the sequence of the vertices visited in the entire graph,
   *  regardless of the connected components of the graph.
   *  @see #reverseTopologicalSort()
   */
  public Vector topologicalSort( );
  
  /**
   *  Perform a reverse topological sort of the entire directed acyclic graph.
   *  Note that the sequence of vertices in the return Vector will not distinguish
   *  between connected components of the graph. This simply calls topologicalSort()
   *  and reverses the sequence of vertices visited.
   *
   *  @return Vector containing the sequence of the vertices visited in the entire graph,
   *  regardless of the connected components of the graph.
   *  @see #topologicalSort()
   */
  public Vector reverseTopologicalSort( );
  
  /**
   * Perform a topological sort of the connected set of a directed acyclic graph
   * to which Vertex startat belongs, starting at Vertex startat.
   *
   * @param	startat	  The Vertex to which you want to start the traversal.
   * 
   * @return  A Vector of vertices in the order that they were visited.
   */
  public Vector topologicalSort( Vertex startat );
  
  /**
   * Perform a reverse topological sort of the connected set of a directed acyclic graph
   * to which Vertex startat belongs, starting at Vertex startat.
   *
   * This method is not part of the GraphTraversal abstract class, but is added
   * here for convenience.
   *
   * @param	startat	  The Vertex to which you want to start the traversal.
   * 
   * @return  A Vector of vertices in the order that they were visited.
   */
  public Vector reverseTopologicalSort( Vertex startat );
}
