/*
 * Copyright 2003 LAAS-CNRS
 *
 *   LAAS-CNRS
 *   7, avenue du Colonel Roche
 *   F-31077 Toulouse Cedex 4 / France
 *
 * Author: Francois Taiani
 *
 * This file is part of OPSBrowser.
 * 
 * OPSBrowser is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * OPSBrowser is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with OPSBrowser; if not, write to the Free Software Foundation,
 * Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 * 
 */
/**
 * MonitorGDB.h
 * 
 * Francois Taiani   <taiani@laas.fr>
 * Time-stamp: <2003-01-29 20:51:38 ftaiani>
 * 
 */
#ifndef MONITORGDB_H
#define MONITORGDB_H

#include <sys/types.h>
#include <regex.h>

#include "Monitor.h"

static char * NULL_STRING = NULL ; 

class MonitorGDB : public Monitor {

public:

  /**
   * Constructor. Launches gdb with the command line arguments
   * "argv[1] ... argv[agrc-1]." (argv[0] is not taken into account.)
   */
  MonitorGDB( int argc =0 , char *  argv[] = &NULL_STRING ) ;

  /**
   * execute a command file into gdb using 
   * the command source + file_name
   */
  void gdbSource( char * fileName ) ;

  /**
   * Tries to identify a breakpoint symbol in the last readlines.
   * If no symbol is indentify the rm_so field of regmatch_t contains
   * -1 and false is returned. (As in the regexec call, see the man pages:
   * 
   * " The regmatch_t structure which is the type of pmatch is defined
   * in regex.h.
   * <pre>
   *    typedef struct
   *    {
   *    regoff_t rm_so;
   *    regoff_t rm_eo;
   *    } regmatch_t;
   * </pre>
   *
   * Each rm_so element that is not -1 indicates the start offset of
   * the next largest substring match within the string.  The relative
   * rm_eo element indicates the end offset of the match. "
   */
  bool identifyBreakpointInLastReadLines(regmatch_t& result) ;

  /**
   * send 'continue' commands to gdb until a breakpoint
   * with a name that match the passed regular expression is met.
   */
  void continueUntillBreakpointOnFunction( char * funcNameRegexp ) ;

  /**
   * Sets breakpoints on all adresses that correspond to the
   * passed symbol. This needs so extra work, because in the
   * strange contexts we work in, the same symbol may denote
   * different addresses. (I should really reads J.R. Levine's
   * book on Linkers and Loaders!).
   * For instance:
   * <pre>
   *   (gdb) info functions ^write$ 
   *   All functions matching regular expression "^write$":
   *
   *   Non-debugging symbols:
   *   0x40025f08  write
   *   0x403f2be0  write
   *   0x4000ddf0  write
   * </pre>
   * The first one libpthread, le 2nd in the libc. I have no idea
   * where the last one is.
   * When using the command breakpoint in gdb, only the lowest
   * address is set with a breakpoint :(
   * <pre>
   *   (gdb) break write
   *   Breakpoint 6 at 0x4000ddf0
   * </pre>
   * This methods tries to work around that limitation by combining
   * calls to 'info functions' and 'break'.
   */
  void setBreakPointSOnSymbol( char * symbol )  ;

  /**
   * Returns true if gdb returned a "The program is not being run."
   * message as output of the last command.
   */
  bool gdbSaidProgramNotBeingRun() ;

  void continueUntilProgramStops() ;

protected:
  static 
  char * createRegularExpForBreakpointDectection(const char * funcNameRegexp) ;

  static
  char * createConcatenatedString( const char * begin  , 
				   const char * middle ,
				   const char * end  ) ;
} ; // EndClass MonitorGDB

#endif // MONITORGDB_H

